!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2009  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief Interface to the message passing library MPI
!> \par History
!>      JGH (02-Jan-2001): New error handling
!>                         Performance tools
!>      JGH (14-Jan-2001): New routines mp_comm_compare, mp_cart_coords, 
!>                                      mp_rank_compare, mp_alltoall
!>      JGH (06-Feb-2001): New routines mp_comm_free
!>      JGH (22-Mar-2001): New routines mp_comm_dup
!>      fawzi (04-NOV-2004): storable performance info (for f77 interface)
!>      Wrapper routine for mpi_gatherv added (22.12.2005,MK)
!>      JGH (13-Feb-2006): Flexibel precision
!>      JGH (15-Feb-2006): single precision mp_alltoall
!> \author JGH
! *****************************************************************************

MODULE message_passing
  USE c_mpi_calls,                     ONLY: mp_alloc_mem,&
                                             mp_free_mem
  USE f77_blas
  USE kinds,                           ONLY: dp,&
                                             dp_size,&
                                             int_8,&
                                             int_size,&
                                             sp
  USE machine,                         ONLY: default_output_unit,&
                                             m_abort,&
                                             m_flush,&
                                             m_walltime

  IMPLICIT NONE
  PRIVATE


#if defined(__mp_timeset__)
  ! Interface to mpi timing routines (wrapper to standard timings routine to work
  ! with circular dependencies)
#include "timings_mp.h"
#endif

#if defined(__parallel)
  INCLUDE "mpif.h"
#if defined (__SGL)
  INTEGER     :: MP_STD_REAL = MPI_REAL
  INTEGER     :: MP_STD_COMPLEX = MPI_COMPLEX
#else
  INTEGER     :: MP_STD_REAL = MPI_DOUBLE_PRECISION
  INTEGER     :: MP_STD_COMPLEX = MPI_DOUBLE_COMPLEX
#endif
#endif

#ifdef __parallel
  LOGICAL, PARAMETER :: cp2k_is_parallel=.TRUE.
  INTEGER, PARAMETER, PUBLIC :: mp_any_tag=MPI_ANY_TAG
  INTEGER, PARAMETER, PUBLIC :: mp_any_source=MPI_ANY_SOURCE
  INTEGER, PARAMETER, PUBLIC :: mp_comm_null=MPI_COMM_NULL
#else
  LOGICAL, PARAMETER :: cp2k_is_parallel=.FALSE.
  INTEGER, PARAMETER, PUBLIC :: mp_any_tag=-1
  INTEGER, PARAMETER, PUBLIC :: mp_any_source=-2
  INTEGER, PARAMETER, PUBLIC :: mp_comm_null=-3
#endif

  CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'message_passing'
  
  ! parameters that might be needed

  ! init and error
  PUBLIC :: mp_world_init, mp_world_finalize
  PUBLIC :: mp_abort

  ! performance gathering
  PUBLIC :: mp_perf_env_type, mp_perf_env_p_type,mp_perf_type
  PUBLIC :: mp_perf_env_create, mp_perf_env_retain, mp_perf_env_release,&
       mp_perf_env_describe
  PUBLIC :: add_mp_perf_env, rm_mp_perf_env, get_mp_perf_env, describe_mp_perf_env

  ! informational / generation of sub comms
  PUBLIC :: mp_environ, mp_comm_compare, mp_cart_coords, mp_rank_compare
  PUBLIC :: mp_cart_create, mp_dims_create, mp_cart_rank, mp_cart_sub, mp_comm_free
  PUBLIC :: mp_comm_dup, mp_comm_split
  PUBLIC :: cp2k_is_parallel

  ! message passing
  PUBLIC :: mp_bcast, mp_sum, mp_max, mp_maxloc, mp_minloc, mp_min, mp_sync
  PUBLIC :: mp_gather, mp_scatter, mp_alltoall, mp_sendrecv, mp_allgather
  PUBLIC :: mp_cart_shift, mp_isend, mp_irecv
  PUBLIC :: mp_sum_scatter, mp_shift, mp_isendrecv, mp_wait, mp_waitall, mp_waitany
  PUBLIC :: mp_gatherv
  PUBLIC :: mp_sum_i8, mp_max_i8
  PUBLIC :: mp_send, mp_recv

  ! RMA
  PUBLIC :: mp_window_type
  PUBLIC :: mp_window_create, mp_window_destroy
  PUBLIC :: mp_window_lock, mp_window_unlock
  PUBLIC :: mp_window_fence
  PUBLIC :: mp_rma_get, mp_rma_put

  ! Memory management
  PUBLIC :: mp_allocate, mp_deallocate


  ! default communicators
  PUBLIC :: MPI_COMM_SELF, MPI_COMM_WORLD, MPI_COMM_NULL
#ifndef __parallel
  INTEGER, PARAMETER :: MPI_COMM_SELF=0,MPI_COMM_WORLD=0, MPI_COMM_NULL=-1
#endif



  !
  ! interfaces to deal easily with scalars / vectors / matrice / ...
  ! of the different types (integers, doubles, logicals, characters)
  !
  INTERFACE mp_minloc
     MODULE PROCEDURE mp_minloc_rv
  END INTERFACE

  INTERFACE mp_maxloc
     MODULE PROCEDURE mp_maxloc_i
  END INTERFACE

  INTERFACE mp_shift
     MODULE PROCEDURE mp_shift_im, mp_shift_i, mp_shift_r1, mp_shift_rm
  END INTERFACE
  
  INTERFACE mp_bcast
     MODULE PROCEDURE mp_bcast_i1, mp_bcast_r1, mp_bcast_c1, mp_bcast_z1, &
          mp_bcast_iv, mp_bcast_rv, mp_bcast_cv, mp_bcast_zv, mp_bcast_l, &
          mp_bcast_lv, mp_bcast_rm, mp_bcast_cm, mp_bcast_im, mp_bcast_rm3
  END INTERFACE

  ! needed in case default integer and int_8 are the same
  INTERFACE mp_sum_i8
     MODULE PROCEDURE mp_sum_8i1, mp_sum_8iv
  END INTERFACE
  INTERFACE mp_max_i8
     MODULE PROCEDURE mp_max_8i1, mp_max_8iv
  END INTERFACE

  INTERFACE mp_sum
     MODULE PROCEDURE mp_sum_i1, mp_sum_r1, mp_sum_c1, mp_sum_iv, &
          mp_sum_rv, mp_sum_cv, mp_sum_im, mp_sum_rm, mp_sum_cm, &
          mp_sum_im3, mp_sum_rm3, mp_sum_cm3, mp_sum_root_rv, &
          mp_sum_root_rm, mp_sum_root_rm3
  END INTERFACE

  INTERFACE mp_max
     MODULE PROCEDURE mp_max_i, mp_max_iv, mp_max_r, mp_max_rv
  END INTERFACE

  INTERFACE mp_min
     MODULE PROCEDURE mp_min_i, mp_min_iv, mp_min_r, mp_min_rv
  END INTERFACE

  INTERFACE mp_gather
     MODULE PROCEDURE mp_gather_i, mp_gather_iv, mp_gather_r, mp_gather_rv
  END INTERFACE
  INTERFACE mp_scatter
     MODULE PROCEDURE mp_scatter_iv
  END INTERFACE

  INTERFACE mp_gatherv
    MODULE PROCEDURE mp_gatherv_iv,mp_gatherv_rv
  END INTERFACE

  INTERFACE mp_allgather
     MODULE PROCEDURE mp_allgather_i1, mp_allgatherv_rv,&
                      mp_allgather_i12, mp_allgather_i23,&
                      mp_allgatherv_iv
  END INTERFACE

  INTERFACE mp_sum_scatter
     MODULE PROCEDURE mp_sum_scatter_rv
  END INTERFACE

  INTERFACE mp_environ
     MODULE PROCEDURE mp_environ_l, mp_environ_c, mp_environ_c2
  END INTERFACE

  INTERFACE mp_alltoall
     MODULE PROCEDURE mp_alltoall_c22v, mp_alltoall_i, mp_alltoall_c22v_sgl, &
                      mp_alltoall_i11v, mp_alltoall_r11v, mp_alltoall_r45,&
                      mp_alltoall_r34,mp_alltoall_r54, mp_alltoall_i8v,&
                      mp_alltoall_z11v, mp_alltoall_i33, mp_alltoall_i22
                      
  END INTERFACE

  INTERFACE mp_send
     MODULE PROCEDURE mp_send_i,mp_send_iv,mp_send_rv
  END INTERFACE

  INTERFACE mp_recv
     MODULE PROCEDURE mp_recv_i,mp_recv_iv,mp_recv_rv
  END INTERFACE

  INTERFACE mp_sendrecv
     MODULE PROCEDURE mp_sendrecv_rm3, mp_sendrecv_rm2, mp_sendrecv_rv
     MODULE PROCEDURE mp_sendrecv_iv
  END INTERFACE

  INTERFACE mp_isendrecv
     MODULE PROCEDURE mp_isendrecv_rm2,mp_isendrecv_rv,mp_isendrecv_8iv
     MODULE PROCEDURE mp_isendrecv_iv
  END INTERFACE

  INTERFACE mp_isend
#if defined(__FFTSGL)
     MODULE PROCEDURE mp_isend_rm3, mp_isend_rm2, mp_isend_cv, mp_isend_rm2_lp, mp_isend_cv_lp
#else
     MODULE PROCEDURE mp_isend_rm3, mp_isend_rm2, mp_isend_cv
#endif
  END INTERFACE

  INTERFACE mp_irecv
#if defined(__FFTSGL)
     MODULE PROCEDURE mp_irecv_rm3, mp_irecv_rm2, mp_irecv_cv, mp_irecv_rm2_lp, mp_irecv_cv_lp
#else
     MODULE PROCEDURE mp_irecv_rm3, mp_irecv_rm2, mp_irecv_cv
#endif
  END INTERFACE

  INTERFACE mp_window_create
     MODULE PROCEDURE mp_win_create_i, mp_win_create_d
  END INTERFACE

  INTERFACE mp_rma_get
     MODULE PROCEDURE mp_rma_get_i, mp_rma_get_d
  END INTERFACE

  INTERFACE mp_rma_put
     MODULE PROCEDURE mp_rma_put_i, mp_rma_put_d
  END INTERFACE

  INTERFACE mp_allocate
     MODULE PROCEDURE mp_allocate_i, mp_allocate_d
  END INTERFACE

  INTERFACE mp_deallocate
     MODULE PROCEDURE mp_deallocate_i, mp_deallocate_d
  END INTERFACE


! *****************************************************************************
!> \brief Windows used for MPI RMA operations
!> \var id          MPI window identifier
!> \var info        MPI window info
! *****************************************************************************
  TYPE mp_window_type
#if defined(__parallel)
     INTEGER        :: id
#else
     INTEGER, DIMENSION(:), POINTER       :: src_i
     REAL(KIND=dp), DIMENSION(:), POINTER :: src_d
#endif
  END TYPE mp_window_type


  ! type internally used to store message passing performance indicators
! *****************************************************************************
  TYPE mp_perf_type
    CHARACTER ( LEN = 20 ) :: name
    INTEGER :: count
    REAL (KIND=dp) :: msg_size
    REAL (KIND=dp) :: time
  END TYPE mp_perf_type

  INTEGER, PARAMETER :: MAX_PERF = 20

! *****************************************************************************
  TYPE mp_perf_env_type
     !private
     INTEGER :: ref_count, id_nr
     TYPE(mp_perf_type), DIMENSION(MAX_PERF) :: mp_perfs
  END TYPE mp_perf_env_type

! *****************************************************************************
  TYPE mp_perf_env_p_type
     TYPE(mp_perf_env_type), POINTER         :: mp_perf_env
  END TYPE mp_perf_env_p_type
     
  ! introduce a stack of mp_perfs, first index is the stack pointer, for convience is replacing
  INTEGER, PARAMETER :: max_stack_size = 10
  INTEGER            :: stack_pointer = 0
  ! target attribute needed as a hack around ifc 7.1 bug 
  TYPE(mp_perf_env_p_type), DIMENSION(max_stack_size), TARGET :: mp_perf_stack

  CHARACTER(LEN=20), PARAMETER :: sname(MAX_PERF) =  &
   (/"MP_Group            ", "MP_Bcast            ", "MP_Allreduce        ", &
     "MP_Gather           ", "MP_Sync             ", "MP_Alltoall         ", &
     "MP_SendRecv         ", "MP_ISendRecv        ", "MP_Wait             ", &
     "MP_comm_split       ", "MP_ISend            ", "MP_IRecv            ", &
     "MP_Send             ", "MP_Recv             ", "MP_Memory           ", &
     "MP_Put              ", "MP_Get              ", "MP_Fence            ", &
     "MP_Window_Lock      ", "MP_Window_Misc      "/)
  REAL(KIND=dp) :: t_start, t_end

  ! we make some assumptions on the length of INTEGERS, REALS and LOGICALS
  INTEGER, PARAMETER :: intlen=BIT_SIZE ( 0 ) / 8, reallen=8, loglen=BIT_SIZE ( 0 ) / 8, charlen=1
  INTEGER, SAVE, PRIVATE :: last_mp_perf_env_id=0

  ! Store the default type for Real and Complex in a special variable

CONTAINS

! *****************************************************************************
!> \brief initializes the system default communicator
!> \param mp_comm [output] : handle of the default communicator
!> \note
!>      should only be called once 
!> \par History
!>      2.2004 created [Joost VandeVondele ]
! *****************************************************************************
  SUBROUTINE mp_world_init(mp_comm)
    INTEGER, INTENT(OUT)                     :: mp_comm
#if defined(__parallel)
    INTEGER                                  :: ierr
    CALL mpi_init ( ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_init @ mp_world_init" )
    CALL mpi_errhandler_set ( MPI_COMM_WORLD, MPI_ERRORS_RETURN, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_errhandler_set @ mp_world_init" )
    mp_comm = MPI_COMM_WORLD
#else
    mp_comm = 0
#endif
  END SUBROUTINE mp_world_init

! *****************************************************************************
!> \brief finalizes the system default communicator
!> \par History
!>      2.2004 created [Joost VandeVondele]
! *****************************************************************************
  SUBROUTINE mp_world_finalize()
#if defined(__parallel)
    INTEGER                                  :: ierr
    CALL mpi_barrier ( MPI_COMM_WORLD,ierr ) ! call mpi directly to avoid 0 stack pointer
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_barrier @ mp_world_finalize" )
    CALL mpi_finalize ( ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_finalize @ mp_world_finalize" )
#endif
  END SUBROUTINE mp_world_finalize

! all the following routines should work for a given communicator, not MPI_WORLD

! *****************************************************************************
!> \brief start and stop the performance indicators
!>      for every call to start there has to be (exactly) one call to stop
!> \param mp_comm the mpi communicator
!> \param scr output unit
!> \note
!>      can be used to measure performance of a sub-part of a program. 
!>      timings measured here will not show up in the outer start/stops
!>      Doesn't need a fresh communicator
!> \par History
!>      2.2004 created [Joost VandeVondele]
! *****************************************************************************
  SUBROUTINE add_mp_perf_env(perf_env)
    TYPE(mp_perf_env_type), OPTIONAL, &
      POINTER                                :: perf_env

    stack_pointer = stack_pointer + 1
    IF (stack_pointer > max_stack_size) THEN
       CALL mp_abort ( "stack_pointer too large : message_passing @ add_mp_perf_env" )   
    ENDIF
    NULLIFY(mp_perf_stack(stack_pointer)%mp_perf_env)
    IF (PRESENT(perf_env)) THEN
       mp_perf_stack(stack_pointer)%mp_perf_env => perf_env
       IF (ASSOCIATED(perf_env)) CALL mp_perf_env_retain(perf_env)
    END IF
    IF (.NOT.ASSOCIATED(mp_perf_stack(stack_pointer)%mp_perf_env)) THEN
       CALL mp_perf_env_create(mp_perf_stack(stack_pointer)%mp_perf_env)
    END IF
  END SUBROUTINE add_mp_perf_env

! *****************************************************************************
  SUBROUTINE mp_perf_env_create(perf_env)
    TYPE(mp_perf_env_type), OPTIONAL, &
      POINTER                                :: perf_env

    INTEGER                                  :: i, stat

    NULLIFY(perf_env)
    ALLOCATE(perf_env,stat=stat)
    IF (stat/=0) THEN
       CALL mp_abort ( "allocation failed in mp_perf_env_create")
    ENDIF
    last_mp_perf_env_id=last_mp_perf_env_id+1
    perf_env%id_nr=last_mp_perf_env_id
    perf_env%ref_count=1
    DO i = 1, MAX_PERF
       perf_env%mp_perfs(i) % name = sname ( i )
       perf_env%mp_perfs(i) % count = 0
       perf_env%mp_perfs(i) % msg_size = 0.0_dp
       perf_env%mp_perfs(i) % time = 0.0_dp
    END DO

  END SUBROUTINE mp_perf_env_create

! *****************************************************************************
  SUBROUTINE mp_perf_env_release(perf_env)
    TYPE(mp_perf_env_type), POINTER          :: perf_env

    INTEGER                                  :: stat

    IF (ASSOCIATED(perf_env)) THEN
       IF (perf_env%ref_count<1) THEN
          CALL mp_abort(&
               "invalid ref_count: message_passing @ mp_perf_env_release")
       END IF
       perf_env%ref_count=perf_env%ref_count-1
       IF (perf_env%ref_count==0) THEN
          DEALLOCATE(perf_env,stat=stat)
          IF (stat/=0) THEN
             CALL mp_abort("deallocation error: message_passing @ mp_perf_env_release")
          END IF
       END IF
    END IF
    NULLIFY(perf_env)
  END SUBROUTINE mp_perf_env_release

! *****************************************************************************
  SUBROUTINE mp_perf_env_retain(perf_env)
    TYPE(mp_perf_env_type), POINTER          :: perf_env

    IF (.NOT.ASSOCIATED(perf_env)) THEN
       CALL mp_abort("unassociated perf_env: message_passing @ mp_perf_env_retain")
    END IF
    IF (perf_env%ref_count<1) THEN
       CALL mp_abort("invalid ref_count: message_passing @ mp_perf_env_retain")
    END IF
    perf_env%ref_count=perf_env%ref_count+1
  END SUBROUTINE mp_perf_env_retain

!.. reports the performance counters for the MPI run
! *****************************************************************************
  SUBROUTINE mp_perf_env_describe ( perf_env, iw )
    TYPE(mp_perf_env_type), POINTER          :: perf_env
    INTEGER, INTENT(IN)                      :: iw

    INTEGER                                  :: i
    REAL(KIND=dp)                            :: per, vol

    IF (.NOT.ASSOCIATED(perf_env)) THEN
       CALL mp_abort ("unassociated perf_env : message_passing @ mp_perf_env_describe" )   
    ENDIF
    IF (perf_env%ref_count<1) THEN
       CALL mp_abort ("invalid perf_env%ref_count : message_passing @ mp_perf_env_describe" )   
    ENDIF
#if defined(__parallel)
    IF ( iw > 0 ) THEN
       WRITE ( iw, '( /, 1X, 79("-") )' )
       WRITE ( iw, '( " -", 77X, "-" )' )
       WRITE ( iw, '( " -", 24X, A, 24X, "-" )' ) ' MESSAGE PASSING PERFORMANCE '
       WRITE ( iw, '( " -", 77X, "-" )' )
       WRITE ( iw, '( 1X, 79("-"), / )' )
       WRITE ( iw, '( A, A, A )' ) ' ROUTINE', '             CALLS ', &
            ' TOT TIME [s]  AVE VOLUME [Bytes]  PERFORMANCE [MB/s]'
       DO i = 1, MAX_PERF

          IF ( perf_env%mp_perfs( i ) % count > 0 ) THEN
             vol = perf_env%mp_perfs( i ) % msg_size / REAL ( perf_env%mp_perfs( i ) % count,KIND=dp)
             IF ( perf_env%mp_perfs( i ) % time > 0.0_dp ) THEN
                per = perf_env%mp_perfs( i ) % msg_size / perf_env%mp_perfs (i) % time * 1.e-6_dp
             ELSE
                per = 0.0_dp
             ENDIF
             IF ( vol < 1.0_dp ) THEN
                WRITE ( iw, '(1X,A15,T17,I10,T27,F14.3)' ) &
                     ADJUSTL ( perf_env%mp_perfs( i ) % name ), perf_env%mp_perfs( i ) % count, &
                     perf_env%mp_perfs( i ) % time
             ELSE
                WRITE ( iw, '(1X,A15,T17,I10,T27,F14.3,T50,F11.0,T69,F12.2)' ) &
                     ADJUSTL ( perf_env%mp_perfs( i ) % name ), perf_env%mp_perfs( i ) % count, &
                     perf_env%mp_perfs( i ) % time, vol, per
             END IF
          ENDIF

       END DO
       WRITE ( iw, '( 1X, 79("-"), / )' )
    END IF
#endif
  END SUBROUTINE mp_perf_env_describe

! *****************************************************************************
  SUBROUTINE rm_mp_perf_env ()
    IF (stack_pointer<1) THEN
       CALL mp_abort ( "no perf_env in the stack : message_passing @ rm_mp_perf_env" )   
    ENDIF
    CALL mp_perf_env_release(mp_perf_stack(stack_pointer)%mp_perf_env)
    stack_pointer = stack_pointer - 1
  END SUBROUTINE rm_mp_perf_env

! *****************************************************************************
  FUNCTION get_mp_perf_env () RESULT(res)
    TYPE(mp_perf_env_type), POINTER          :: res

    IF (stack_pointer<1) THEN
       CALL mp_abort ( "no perf_env in the stack : message_passing @ get_mp_perf_env" )   
    ENDIF
    res => mp_perf_stack(stack_pointer)%mp_perf_env
  END FUNCTION get_mp_perf_env

! *****************************************************************************
  SUBROUTINE describe_mp_perf_env(scr)
    INTEGER, INTENT(in)                      :: scr

    TYPE(mp_perf_env_type), POINTER          :: perf_env

    perf_env => get_mp_perf_env()
    CALL mp_perf_env_describe(perf_env, scr)
  END SUBROUTINE describe_mp_perf_env

! *****************************************************************************
!> \brief adds the performance informations of one call
!> \param error variable to control error logging, stopping,... 
!>        see module cp_error_handling 
!> \author fawzi
! *****************************************************************************
  SUBROUTINE add_perf(perf_id,count,time,msg_size)
    INTEGER, INTENT(in)                      :: perf_id
    INTEGER, INTENT(in), OPTIONAL            :: count
    REAL(KIND=dp), INTENT(in), OPTIONAL      :: time
    INTEGER, INTENT(in), OPTIONAL            :: msg_size

    TYPE(mp_perf_type), POINTER              :: mp_perf

#if defined(__parallel)
    mp_perf => mp_perf_stack (stack_pointer)%mp_perf_env%mp_perfs( perf_id )
    IF (PRESENT(count)) THEN
       mp_perf%count = mp_perf%count + count
    END IF
    IF (PRESENT(time)) THEN
       mp_perf%time = mp_perf%time + time
    END IF
    IF (PRESENT(msg_size)) THEN
       mp_perf%msg_size = mp_perf%msg_size+REAL(msg_size,dp)
    END IF
#endif

  END SUBROUTINE add_perf

! *****************************************************************************
!> \brief globally stops all tasks, can optionally print a message.
!>       this is intended to be low level, most of CP2K should rather use cp_error_handling
! *****************************************************************************
  SUBROUTINE mp_abort (message)
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: message

    INTEGER                                  :: ierr, numtask, taskid

#if defined(__parallel)
    CALL mp_environ ( numtask, taskid, MPI_COMM_WORLD )
#else
    numtask=1
    taskid=0
#endif
    IF (PRESENT(message)) THEN
       WRITE(default_output_unit,'(A)')         ' CP2K| '//TRIM(message)
    ENDIF
    WRITE(default_output_unit,'(A,I0)')   ' CP2K| Abnormal program termination, stopped by process number ',taskid
    CALL m_flush(default_output_unit)

#if defined(__parallel)
    CALL mpi_abort ( MPI_COMM_WORLD, 1, ierr )
#else
    CALL m_abort() ! uncomment if you want nice core dumps
#endif
    ! this routine never returns
    STOP 1
  END SUBROUTINE mp_abort

! *****************************************************************************
!> \brief stops *after an mpi error* translating the error code
!> \param ierr an error code * returned by an mpi call *
!> \note
!>       this function is private to message_passing.F
! *****************************************************************************
  SUBROUTINE mp_stop ( ierr, prg_code )
    INTEGER, INTENT(IN)                      :: ierr
    CHARACTER(LEN=*)                         :: prg_code

    INTEGER                                  :: istat, len
    CHARACTER(LEN=512)                       :: full_error
#if defined(__parallel)
    CHARACTER ( LEN = MPI_MAX_ERROR_STRING ) :: error_string
#else
    CHARACTER ( LEN = 512 )                  :: error_string
#endif

#if defined(__parallel)
    CALL mpi_error_string ( ierr, error_string, len, istat )
    WRITE(full_error,'(A,I0,A)') ' MPI error ',ierr,' in '//TRIM(prg_code)//' : '//error_string(1:len)
#else
    WRITE(full_error,'(A,I0,A)') ' MPI error (!?) ',ierr,' in '//TRIM(prg_code)
#endif

    CALL mp_abort(full_error)

  END SUBROUTINE mp_stop

! *****************************************************************************
!> \brief synchronizes with a barrier a given group of mpi tasks
!> \param group mpi communicator
! *****************************************************************************
  SUBROUTINE mp_sync ( group )
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sync', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_barrier ( group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_barrier @ mp_sync" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=5,count=1,time=t_end-t_start)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_sync

! *****************************************************************************
!> \brief returns number of tasks and task id for a given mpi group
!>       simple and cartesian version
!> \param groupid mpi communicator
!> \note
!>         ..mp_world_setup is gone, use mp_environ instead (i.e. give a groupid explicitly)
! *****************************************************************************
  SUBROUTINE mp_environ_l ( numtask, taskid, groupid )

    INTEGER, INTENT(OUT)                     :: numtask, taskid
    INTEGER, INTENT(IN)                      :: groupid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_environ_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr    = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    numtask = 1
    taskid  = 0
#if defined(__parallel)
    CALL mpi_comm_rank ( groupid, taskid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_rank @ mp_environ_l" )

    CALL mpi_comm_size ( groupid, numtask, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_environ_l" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_environ_l

! *****************************************************************************
  SUBROUTINE mp_environ_c ( numtask, dims, task_coor, groupid )

    INTEGER, INTENT(OUT)                     :: numtask, dims( 2 ), &
                                                task_coor( 2 )
    INTEGER, INTENT(IN)                      :: groupid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_environ_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    LOGICAL                                  :: periods( 2 )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    numtask = 1
    task_coor = 0
    dims = 1
#if defined(__parallel)
    CALL mpi_comm_size ( groupid, numtask, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_environ_c" )

    CALL mpi_cart_get ( groupid, 2, dims, periods, task_coor, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_cart_get @ mp_environ_c" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_environ_c

! *****************************************************************************
  SUBROUTINE mp_environ_c2 ( comm, ndims, dims, task_coor, periods )

    INTEGER, INTENT(IN)                      :: comm, ndims
    INTEGER, INTENT(OUT)                     :: dims( ndims ), &
                                                task_coor( ndims )
    LOGICAL, INTENT(out)                     :: periods( ndims )

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_environ_c2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    task_coor = 0
    dims = 1
    periods=.FALSE.
#if defined(__parallel)
    CALL mpi_cart_get ( comm, ndims, dims, periods, task_coor, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_cart_get @ mp_environ_c" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_environ_c2

!..mp_cart_create
! *****************************************************************************
  SUBROUTINE mp_cart_create ( comm_old, ndims, dims, pos, comm_cart )

    INTEGER, INTENT(IN)                      :: comm_old, ndims
    INTEGER, INTENT(INOUT)                   :: dims( : )
    INTEGER, INTENT(OUT)                     :: pos( : ), comm_cart

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_create', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, nodes
    LOGICAL                                  :: period(1:ndims), reorder

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    pos ( 1:ndims ) = -1
    comm_cart = comm_old
#if defined(__parallel)

    t_start = m_walltime ( )
    CALL mpi_comm_size ( comm_old, nodes, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_cart_create" )

    IF (ANY(dims == 0)) CALL mpi_dims_create(nodes,ndims,dims,ierr)
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_dims_create @ mp_cart_create" )

    ! FIX ME.  Quick hack to avoid problems with realspace grids for compilers
    ! like IBM that actually reorder the processors when creating the new
    ! communicator
    reorder = .FALSE.
    period = .TRUE.
    CALL mpi_cart_create ( comm_old, ndims, dims, period, reorder, comm_cart, &
         ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_cart_create @ mp_cart_create" )

    IF (comm_cart /= MPI_COMM_NULL) THEN
       CALL mpi_cart_get ( comm_cart, ndims, dims, period, pos, ierr )
       IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_cart_get @ mp_cart_create" )
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=1,count=1)
    CALL add_perf(perf_id=2,time=t_end-t_start)
#else
    pos ( 1:ndims ) = 0
    dims = 1
    comm_cart = 0
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_cart_create

!..mp_cart_coords 
! *****************************************************************************
  SUBROUTINE mp_cart_coords ( comm, rank, coords)

    INTEGER, INTENT(IN)                      :: comm, rank
    INTEGER, DIMENSION(:), INTENT(OUT)       :: coords

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_coords', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    m = SIZE ( coords )
#if defined(__parallel)
    CALL mpi_cart_coords ( comm, rank, m, coords, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_cart_coords @ mp_cart_coords" )
#else
    coords = 0
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_cart_coords

!..mp_cart_shift
! *****************************************************************************
  SUBROUTINE mp_cart_shift ( comm, dir, disp, source, dest )

    INTEGER, INTENT(IN)                      :: comm, dir, disp
    INTEGER, INTENT(OUT)                     :: source, dest

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_shift', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    CALL mpi_cart_shift ( comm, dir, disp, source, dest, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_cart_shift @ mp_cart_shift" )
#else
    source = 0
    dest = 0
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_cart_shift

!..mp_comm_compare
! *****************************************************************************
  SUBROUTINE mp_comm_compare ( comm1, comm2, RESULT)

    INTEGER, INTENT(IN)                      :: comm1, comm2
    INTEGER, INTENT(OUT)                     :: RESULT

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_compare', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, iout

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    RESULT = 0
#if defined(__parallel)
    CALL mpi_comm_compare ( comm1, comm2, iout, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_compare @ mp_comm_compare" )
    SELECT CASE(iout)
    CASE(MPI_IDENT)
       RESULT=0
    CASE(MPI_CONGRUENT)
       RESULT=1
    CASE(MPI_SIMILAR)
       RESULT=2
    CASE (MPI_UNEQUAL)
       RESULT=3
    CASE default
       RESULT=4
    END SELECT
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_comm_compare

!..mp_cart_sub
! *****************************************************************************
  SUBROUTINE mp_cart_sub ( comm, rdim, sub_comm )

    INTEGER, INTENT(IN)                      :: comm
    LOGICAL, DIMENSION(:), INTENT(IN)        :: rdim
    INTEGER, INTENT(OUT)                     :: sub_comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_sub', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    sub_comm = 0
#if defined(__parallel)
    CALL mpi_cart_sub ( comm, rdim, sub_comm, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_cart_sub @ mp_cart_sub" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_cart_sub

!..mp_comm_free
! *****************************************************************************
  SUBROUTINE mp_comm_free ( comm )

    INTEGER, INTENT(IN)                      :: comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_free', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    CALL mpi_comm_free ( comm, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_free @ mp_comm_free" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_comm_free

!..mp_comm_dup
! *****************************************************************************
  SUBROUTINE mp_comm_dup ( comm1, comm2 )

    INTEGER, INTENT(IN)                      :: comm1
    INTEGER, INTENT(OUT)                     :: comm2

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_dup', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    CALL mpi_comm_dup ( comm1, comm2, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_dup @ mp_comm_dup" )
#else
    comm2=comm1
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_comm_dup

!..mp_rank_compare
! *****************************************************************************
  SUBROUTINE mp_rank_compare ( comm1, comm2, rank )

    INTEGER, INTENT(IN)                      :: comm1, comm2
    INTEGER, DIMENSION(:), INTENT(OUT)       :: rank

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rank_compare', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: g1, g2, handle, i, ierr, n, &
                                                n1, n2
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: rin

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    rank = 0
#if defined(__parallel)
    CALL mpi_comm_size ( comm1, n1, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_size @ mp_rank_compare" )
    CALL mpi_comm_size ( comm2, n2, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_size @ mp_rank_compare" )
    n = MAX ( n1, n2 )
    CALL mpi_comm_group ( comm1, g1, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_group @ mp_rank_compare" )
    CALL mpi_comm_group ( comm2, g2, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_group @ mp_rank_compare" )
    ALLOCATE ( rin ( 0 : n - 1 ), STAT = ierr )
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_rank_compare" )
    DO i = 0, n-1
       rin ( i ) = i
    END DO
    CALL mpi_group_translate_ranks ( g1, n, rin, g2, rank, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, &
         "mpi_group_translate_rank @ mp_rank_compare" )
    DEALLOCATE ( rin, STAT = ierr )
    IF ( ierr /= 0 ) CALL mp_abort( "deallocate @ mp_rank_compare" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_rank_compare

!..mp_dims_create
! *****************************************************************************
  SUBROUTINE mp_dims_create ( nodes, dims )

    INTEGER, INTENT(IN)                      :: nodes
    INTEGER, DIMENSION(:), INTENT(INOUT)     :: dims

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_dims_create', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, ndim

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    ndim = SIZE ( dims )
    IF (ANY(dims == 0)) CALL mpi_dims_create(nodes,ndim,dims,ierr)
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_dims_create @ mp_dims_create" )
#else
    dims = 1
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_dims_create

! this routine will shift around the data in msg
! i.e. msg will be moved from rank to rank+displ_in (in a circular way)
! displ_in will be 1 by default (others not tested)
! the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_im( msg, group, displ_in)

    INTEGER, INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: displ, handle, ierror, left, &
                                                msglen, myrank, nprocs, &
                                                right, tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierror = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    CALL mpi_comm_rank(group,myrank,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_rank @ mp_shift" )
    CALL mpi_comm_size(group,nprocs,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_size @ mp_shift" )
    IF (PRESENT(displ_in)) THEN
       displ=displ_in
    ELSE
       displ=1
    ENDIF
    right=MODULO(myrank+displ,nprocs)
    left =MODULO(myrank-displ,nprocs)
    tag=17
    msglen = SIZE(msg)
    t_start = m_walltime ( )
    CALL mpi_sendrecv_replace(msg,msglen,MPI_INTEGER,right,tag,left,tag, &
         group,status(1),ierror)
    t_end = m_walltime ( )
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_sendrecv_replace @ mp_shift_im" )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,msg_size=msglen*intlen)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_shift_im

! this routine will shift around the data in msg
! i.e. msg will be moved from rank to rank+displ_in (in a circular way)
! displ_in will be 1 by default (others not tested)
! the message array needs to be the same size on all processes
! *****************************************************************************
  SUBROUTINE mp_shift_i( msg, group, displ_in)

    INTEGER, INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: displ, handle, ierror, left, &
                                                msglen, myrank, nprocs, &
                                                right, tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierror = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    CALL mpi_comm_rank(group,myrank,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_rank @ mp_shift" )
    CALL mpi_comm_size(group,nprocs,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_size @ mp_shift" )
    IF (PRESENT(displ_in)) THEN
       displ=displ_in
    ELSE
       displ=1
    ENDIF
    right=MODULO(myrank+displ,nprocs)
    left =MODULO(myrank-displ,nprocs)
    tag=17
    msglen = SIZE(msg)
    t_start = m_walltime ( )
    CALL mpi_sendrecv_replace(msg,msglen,MPI_INTEGER,right,tag,left,tag, &
         group,status(1),ierror)
    t_end = m_walltime ( )
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_sendrecv_replace @ mp_shift_im" )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,msg_size=msglen*intlen)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_shift_i

! *****************************************************************************
  SUBROUTINE mp_shift_rm( msg, group, displ_in)

    REAL(KIND=dp), INTENT(INOUT)             :: msg( :, : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: displ, handle, ierror, left, &
                                                msglen, myrank, nprocs, &
                                                right, tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierror = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    CALL mpi_comm_rank(group,myrank,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_rank @ mp_shift" )
    CALL mpi_comm_size(group,nprocs,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_size @ mp_shift" )
    IF (PRESENT(displ_in)) THEN
       displ=displ_in 
    ELSE
       displ=1
    ENDIF
    right=MODULO(myrank+displ,nprocs) 
    left =MODULO(myrank-displ,nprocs) 
    tag=18
    msglen = SIZE(msg)
    t_start = m_walltime ( )
    CALL mpi_sendrecv_replace(msg,msglen,MP_STD_REAL,right,tag,left,&
         tag,group,status(1),ierror)
    t_end = m_walltime ( )
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_sendrecv_replace @ mp_shift_rm" )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,msg_size=msglen*reallen)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_shift_rm

! *****************************************************************************
  SUBROUTINE mp_shift_r1( msg, group, displ_in)

    REAL(KIND=dp), INTENT(INOUT)             :: msg( : )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, INTENT(IN), OPTIONAL            :: displ_in

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_shift_r1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: displ, handle, ierror, left, &
                                                msglen, myrank, nprocs, &
                                                right, tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierror = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    CALL mpi_comm_rank(group,myrank,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_rank @ mp_shift" )
    CALL mpi_comm_size(group,nprocs,ierror)
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_comm_size @ mp_shift" )
    IF (PRESENT(displ_in)) THEN
       displ=displ_in
    ELSE
       displ=1
    ENDIF
    right=MODULO(myrank+displ,nprocs)
    left =MODULO(myrank-displ,nprocs)
    tag=19
    msglen = SIZE(msg)
    t_start = m_walltime ( )
    CALL mpi_sendrecv_replace(msg,msglen,MP_STD_REAL,right,tag,left,&
         tag,group,status(1),ierror)
    t_end = m_walltime ( )
    IF ( ierror /= 0 ) CALL mp_stop ( ierror, "mpi_sendrecv_replace @ mp_shift_r1" )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,msg_size=msglen*reallen)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_shift_r1

!..mp_cart_rank
! *****************************************************************************
  SUBROUTINE mp_cart_rank ( group, pos, rank )
    INTEGER, INTENT(IN)                      :: group
    INTEGER, DIMENSION(:), INTENT(IN)        :: pos
    INTEGER, INTENT(OUT)                     :: rank

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_cart_rank', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    CALL mpi_cart_rank ( group, pos, rank, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_cart_rank @ mp_cart_rank" )
#else
    rank = 0
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_cart_rank
!..mp_alltoall

! *****************************************************************************
  SUBROUTINE mp_alltoall_i11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    INTEGER, DIMENSION(:), INTENT(IN)        :: sb, scount, sdispl
    INTEGER, DIMENSION(:), INTENT(INOUT)     :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    ierr = 0
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoallv ( sb, scount, sdispl, MPI_INTEGER, &
         rb, rcount, rdispl, MPI_INTEGER, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoallv @ mp_alltoall_i11v" )
    t_end = m_walltime ( )
    msglen = SUM ( scount ) + SUM ( rcount )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    rb(rdispl(1)+1:rdispl(1)+rcount(1))=sb(sdispl(1)+1:sdispl(1)+scount(1))
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_i11v

! *****************************************************************************
  SUBROUTINE mp_alltoall_i8v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    INTEGER(kind=int_8), DIMENSION(:), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    INTEGER(kind=int_8), DIMENSION(:), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i8v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, int_8_size_4, &
                                                msglen
    INTEGER(kind=int_8)                      :: dummy, int_8_size

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    dummy = 10
    int_8_size = BIT_SIZE (dummy) / 8
    int_8_size_4 = int_8_size

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoallv ( sb, scount, sdispl, MPI_INTEGER8, &
         rb, rcount, rdispl, MPI_INTEGER8, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoallv @ mp_alltoall_i8v" )
    t_end = m_walltime ( )
    msglen = SUM ( scount ) + SUM ( rcount )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*int_8_size_4)
#else
    rb(rdispl(1)+1:rdispl(1)+rcount(1))=sb(sdispl(1)+1:sdispl(1)+scount(1))
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_i8v

! *****************************************************************************
  SUBROUTINE mp_alltoall_z11v ( sb, scount, sdispl, rb, rcount, rdispl,len_string, group )

    INTEGER, INTENT(IN)                      :: len_string
    INTEGER, DIMENSION(:), INTENT(IN)        :: rdispl, rcount
    CHARACTER(LEN=len_string), &
      DIMENSION(:), INTENT(INOUT)            :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: sdispl, scount
    CHARACTER(LEN=len_string), &
      DIMENSION(:), INTENT(IN)               :: sb
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_z11v', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(SIZE(scount))         :: len_r, len_s
    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    len_s(:)=scount(:)*len_string
    len_r(:)=rcount(:)*len_string  
    CALL mpi_alltoallv ( sb, len_s, sdispl*len_string, MPI_CHARACTER, &
         rb, len_r, rdispl*len_string, MPI_CHARACTER, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoallv @ mp_alltoall_z11v" )
    msglen = SUM ( scount ) + SUM ( rcount )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    rb(rdispl(1)+1:rdispl(1)+rcount(1))=sb(sdispl(1)+1:sdispl(1)+scount(1))
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_z11v

! *****************************************************************************
  SUBROUTINE mp_alltoall_r11v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    REAL(KIND=dp), DIMENSION(:), INTENT(IN)  :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    REAL(KIND=dp), DIMENSION(:), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r11v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoallv ( sb, scount, sdispl, MP_STD_REAL, &
         rb, rcount, rdispl, MP_STD_REAL, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoallv @ mp_alltoall_r11v" )
    msglen = SUM ( scount ) + SUM ( rcount )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    rb(rdispl(1)+1:rdispl(1)+rcount(1))=sb(sdispl(1)+1:sdispl(1)+scount(1))
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_r11v

!..mp_alltoall
! *****************************************************************************
  SUBROUTINE mp_alltoall_c22v ( sb, scount, sdispl, rb, rcount, rdispl, group )

    COMPLEX(KIND=dp), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(KIND=dp), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c22v', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoallv ( sb, scount, sdispl, MP_STD_COMPLEX, &
         rb, rcount, rdispl, MP_STD_COMPLEX, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoallv @ mp_alltoall_c22v" )
    msglen = SUM ( scount ) + SUM ( rcount )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#else
    rb=sb
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_c22v

#if defined(__SGL)
! *****************************************************************************
  SUBROUTINE mp_alltoall_c22v_sgl ( dummy )
    LOGICAL, INTENT(IN)                      :: dummy

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c22v_sgl', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    CALL mp_abort("illegal call !? to mp_alltoall_c22v_sgl" )
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_alltoall_c22v_sgl
#else
! *****************************************************************************
  SUBROUTINE mp_alltoall_c22v_sgl ( sb, scount, sdispl, rb, rcount, rdispl, group )

    COMPLEX(KIND=sp), DIMENSION(:, :), &
      INTENT(IN)                             :: sb
    INTEGER, DIMENSION(:), INTENT(IN)        :: scount, sdispl
    COMPLEX(KIND=sp), DIMENSION(:, :), &
      INTENT(INOUT)                          :: rb
    INTEGER, DIMENSION(:), INTENT(IN)        :: rcount, rdispl
    INTEGER, INTENT(IN)                      :: group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_c22v_sgl', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoallv ( sb, scount, sdispl, MPI_COMPLEX, &
         rb, rcount, rdispl, MPI_COMPLEX, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoallv @ mp_alltoall_c22v" )
    msglen = SUM ( scount ) + SUM ( rcount )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    rb=sb
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_c22v_sgl
#endif

! *****************************************************************************
!> \brief All-to-all scatter gather, 3-D arrays
!> \par The first two indices specify the data while the last index counts
!>      the processes
!> \param[in] sb    array with data to send
!> \param[out] rb   array into which data is received
!> \param[in] count  number of elements to send/receive (product of the
!>                   extents of the first two dimensions)
! *****************************************************************************
  SUBROUTINE mp_alltoall_i33 ( sb, rb, count, group )

    INTEGER, DIMENSION(:, :, :), INTENT(IN)  :: sb
    INTEGER, DIMENSION(:, :, :), INTENT(OUT) :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i33', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, np

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoall ( sb, count, MPI_INTEGER, &
         rb, count, MPI_INTEGER, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoall @ mp_alltoall_i33" )
    CALL mpi_comm_size ( group, np, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_alltoall_i33" )
    msglen = 2 * count * np
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    rb=sb
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_i33



! *****************************************************************************
  SUBROUTINE mp_alltoall_i ( sb, rb, count, group )

    INTEGER, DIMENSION(:), INTENT(IN)        :: sb
    INTEGER, DIMENSION(:), INTENT(OUT)       :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, np

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoall ( sb, count, MPI_INTEGER, &
         rb, count, MPI_INTEGER, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoall @ mp_alltoall_i" )
    CALL mpi_comm_size ( group, np, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_alltoall_i" )
    msglen = 2 * count * np
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    rb=sb
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_i


! *****************************************************************************
  SUBROUTINE mp_alltoall_i22 ( sb, rb, count, group )

    INTEGER, DIMENSION(:, :), INTENT(IN)     :: sb
    INTEGER, DIMENSION(:, :), INTENT(OUT)    :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_i22', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, np

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoall ( sb, count, MPI_INTEGER, &
         rb, count, MPI_INTEGER, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoall @ mp_alltoall_i22" )
    CALL mpi_comm_size ( group, np, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_alltoall_i22" )
    msglen = 2 * SIZE(sb) * np
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    rb=sb
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_i22


! *****************************************************************************
  SUBROUTINE mp_alltoall_r45 ( sb, rb, count, group )

    REAL(kind=dp), DIMENSION(:, :, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=dp), &
      DIMENSION(:, :, :, :, :), INTENT(OUT)  :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r45', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, np

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoall ( sb, count, MP_STD_REAL, &
         rb, count, MP_STD_REAL, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoall @ mp_alltoall_r" )
    CALL mpi_comm_size ( group, np, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_alltoall_r" )
    msglen = 2 * count * np
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_r45

! *****************************************************************************
  SUBROUTINE mp_alltoall_r34 ( sb, rb, count, group )

    REAL(kind=dp), DIMENSION(:, :, :), &
      INTENT(IN)                             :: sb
    REAL(kind=dp), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r34', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, np

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoall ( sb, count, MP_STD_REAL, &
         rb, count, MP_STD_REAL, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoall @ mp_alltoall_r" )
    CALL mpi_comm_size ( group, np, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_alltoall_r" )
    msglen = 2 * count * np
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_r34

! *****************************************************************************
  SUBROUTINE mp_alltoall_r54 ( sb, rb, count, group )

    REAL(kind=dp), &
      DIMENSION(:, :, :, :, :), INTENT(IN)   :: sb
    REAL(kind=dp), DIMENSION(:, :, :, :), &
      INTENT(OUT)                            :: rb
    INTEGER, INTENT(IN)                      :: count, group

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_alltoall_r54', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, np

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_alltoall ( sb, count, MP_STD_REAL, &
         rb, count, MP_STD_REAL, group, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_alltoall @ mp_alltoall_r" )
    CALL mpi_comm_size ( group, np, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_size @ mp_alltoall_r" )
    msglen = 2 * count * np
    t_end = m_walltime ( )
    CALL add_perf(perf_id=6,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_alltoall_r54

! ..mp_send
! *****************************************************************************
  SUBROUTINE mp_send_iv(msg,dest,tag,gid)
    INTEGER                                  :: msg( : ), dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_send(msg,msglen,MPI_INTEGER,dest,tag,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_send @ mp_send_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=13,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_send_iv

! *****************************************************************************
  SUBROUTINE mp_send_rv(msg,dest,tag,gid)
    REAL(KIND=dp)                            :: msg( : )
    INTEGER                                  :: dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_send(msg,msglen,MP_STD_REAL,dest,tag,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_send @ mp_send_rv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=13,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_send_rv

! *****************************************************************************
  SUBROUTINE mp_send_i(msg,dest,tag,gid)
    INTEGER                                  :: msg, dest, tag, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_send_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_send(msg,msglen,MPI_INTEGER,dest,tag,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_send @ mp_send_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=13,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_send_i

! ..mp_recv
! *****************************************************************************
  SUBROUTINE mp_recv_iv(msg,source,tag,gid)
    INTEGER, INTENT(INOUT)                   :: msg( : ), source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )
    CALL mpi_recv(msg,msglen,MPI_INTEGER,source,tag,gid,status,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_recv @ mp_recv_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=14,count=1,time=t_end-t_start,msg_size=msglen*intlen)
    source = status(MPI_SOURCE)
    tag = status(MPI_TAG)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_recv_iv

! *****************************************************************************
  SUBROUTINE mp_recv_rv(msg,source,tag,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( : )
    INTEGER, INTENT(INOUT)                   :: source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )
    CALL mpi_recv(msg,msglen,MP_STD_REAL,source,tag,gid,status,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_recv @ mp_recv_rv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=14,count=1,time=t_end-t_start,msg_size=msglen*intlen)
    source = status(MPI_SOURCE)
    tag = status(MPI_TAG)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_recv_rv

! *****************************************************************************
  SUBROUTINE mp_recv_i(msg,source,tag,gid)
    INTEGER, INTENT(INOUT)                   :: msg, source, tag
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_recv_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )
    CALL mpi_recv(msg,msglen,MPI_INTEGER,source,tag,gid,status,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_recv @ mp_recv_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=14,count=1,time=t_end-t_start,msg_size=msglen*intlen)
    source = status(MPI_SOURCE)
    tag = status(MPI_TAG)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_recv_i

!..mp_bcast
! *****************************************************************************
  SUBROUTINE mp_bcast_i1(msg,source,gid)
    INTEGER                                  :: msg, source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_i1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MPI_INTEGER,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_i1" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_i1
! *****************************************************************************
  SUBROUTINE mp_bcast_iv(msg,source,gid)
    INTEGER                                  :: msg( : ), source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MPI_INTEGER,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_iv
! *****************************************************************************
  SUBROUTINE mp_bcast_im(msg,source,gid)
    INTEGER                                  :: msg( :, : ), source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MPI_INTEGER,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_im" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_im
! *****************************************************************************
  SUBROUTINE mp_bcast_r1(msg,source,gid)
    REAL(KIND=dp)                            :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_r1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MP_STD_REAL,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_r1" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_r1
! *****************************************************************************
  SUBROUTINE mp_bcast_rv(msg,source,gid)
    REAL(KIND=dp)                            :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MP_STD_REAL,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_rv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_rv
! *****************************************************************************
  SUBROUTINE mp_bcast_rm(msg,source,gid)
    REAL(KIND=dp)                            :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MP_STD_REAL,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_rm" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_rm
! *****************************************************************************
  SUBROUTINE mp_bcast_rm3(msg,source,gid)
    REAL(KIND=dp)                            :: msg( :, :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MP_STD_REAL,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_rm" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_rm3
! *****************************************************************************
  SUBROUTINE mp_bcast_c1(msg,source,gid)
    COMPLEX(KIND=dp)                         :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_c1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MP_STD_COMPLEX,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_c1" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_c1
! *****************************************************************************
  SUBROUTINE mp_bcast_cv(msg,source,gid)
    COMPLEX(KIND=dp)                         :: msg( : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MP_STD_COMPLEX,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_cv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_cv
! *****************************************************************************
  SUBROUTINE mp_bcast_cm(msg,source,gid)
    COMPLEX(KIND=dp)                         :: msg( :, : )
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_cm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MP_STD_COMPLEX,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_cm")
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_cm
! *****************************************************************************
  SUBROUTINE mp_bcast_l(msg,source,gid)
    LOGICAL                                  :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MPI_LOGICAL,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_l" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*loglen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_l
! *****************************************************************************
  SUBROUTINE mp_bcast_lv(msg,source,gid)
    LOGICAL                                  :: msg(:)
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_lv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = SIZE(msg)
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_bcast(msg,msglen,MPI_LOGICAL,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_l" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*loglen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_lv
! *****************************************************************************
  SUBROUTINE mp_bcast_z1(msg,source,gid)
    CHARACTER(LEN=*)                         :: msg
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_z1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, i, ierr, msglen
    INTEGER, ALLOCATABLE                     :: imsg( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = LEN_TRIM(msg)
    CALL mp_bcast_i1(msglen, source, gid)
    ! this is a workaround to avoid problems on the T3E
    ! at the moment we have a data alignment error when trying to
    ! broadcats characters on the T3E (not always!)
    ! JH 19/3/99 on galileo
    ! CALL mpi_bcast(msg,msglen,MPI_CHARACTER,source,gid,ierr)
    ALLOCATE (imsg(1:msglen))
    DO i = 1, msglen
       imsg(i) = ICHAR(msg(i:i))
    END DO
    CALL mpi_bcast(imsg,msglen,MPI_INTEGER,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_z" )
    msg = ""
    DO i = 1, msglen
       msg(i:i) = CHAR(imsg(i))
    END DO
    DEALLOCATE (imsg)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*charlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_z1
! *****************************************************************************
  SUBROUTINE mp_bcast_zv(msg,source,gid)
    CHARACTER(LEN=*)                         :: msg(:)
    INTEGER                                  :: source, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_bcast_zv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, i, ierr, j, k, &
                                                msglen, msgsiz
    INTEGER, ALLOCATABLE                     :: imsg(:), imsglen(:)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msgsiz = SIZE(msg)
    ! Determine size of the minimum array of integers to bradcast the string
    ALLOCATE (imsglen(1:msgsiz))
    DO j = 1, msgsiz
       imsglen(j) = LEN_TRIM(msg(j))
    END DO
    CALL mp_bcast_iv(imsglen, source, gid)
    msglen = SUM(imsglen)
    ! this is a workaround to avoid problems on the T3E
    ! at the moment we have a data alignment error when trying to
    ! broadcats characters on the T3E (not always!)
    ! JH 19/3/99 on galileo
    ! CALL mpi_bcast(msg,msglen,MPI_CHARACTER,source,gid,ierr)
    ALLOCATE (imsg(1:msglen))
    k = 0
    DO j = 1, msgsiz
       DO i = 1, imsglen(j)
          k = k + 1
          imsg(k) = ICHAR(msg(j)(i:i))
       END DO
    END DO
    CALL mpi_bcast(imsg,msglen,MPI_INTEGER,source,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_bcast @ mp_bcast_z" )
    msg = ""
    k = 0
    DO j = 1, msgsiz
       DO i = 1, imsglen(j)
          k = k + 1
          msg(j)(i:i) = CHAR(imsg(k))
       END DO
    END DO
    DEALLOCATE (imsg)
    DEALLOCATE (imsglen)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=2,count=1,time=t_end-t_start,msg_size=msglen*charlen*msgsiz)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_bcast_zv

!..mp_sum
! *****************************************************************************
  SUBROUTINE mp_sum_i1(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_i1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_SUM,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_i1" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_i1

! *****************************************************************************
  SUBROUTINE mp_sum_8i1(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)       :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_8i1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER(KIND=int_8)                      :: res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER8,MPI_SUM,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_8i1" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_8i1

! *****************************************************************************
  SUBROUTINE mp_max_8i1(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)       :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_8i1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER(KIND=int_8)                      :: res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER8,MPI_MAX,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_max_8i1" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_max_8i1


! *****************************************************************************
  SUBROUTINE mp_sum_iv(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE                     :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_iv" )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_SUM,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_iv" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_iv

! *****************************************************************************
  SUBROUTINE mp_sum_8iv(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)       :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_8iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER(KIND=int_8), ALLOCATABLE         :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_iv" )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER8,MPI_SUM,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_8iv" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_8iv

! *****************************************************************************
  SUBROUTINE mp_sum_im(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_im', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m1, m2, msglen
    INTEGER, ALLOCATABLE                     :: res( :, : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    ALLOCATE (res(m1,m2),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_im" )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_SUM,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_im" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_im
! *****************************************************************************
  SUBROUTINE mp_sum_im3(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_im3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m1, m2, m3, &
                                                msglen
    INTEGER, ALLOCATABLE                     :: res( :, :, : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    m3 = SIZE(msg,3)
    ALLOCATE (res(m1,m2,m3),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_im3" )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_SUM,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_im3" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_im3
! *****************************************************************************
  SUBROUTINE mp_sum_r1(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_r1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    REAL(KIND=dp)                            :: res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 1
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_r1" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_r1
! *****************************************************************************
  SUBROUTINE mp_sum_rv(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    REAL(KIND=dp), ALLOCATABLE               :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_rv" )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_rv" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_rv
! *****************************************************************************
  SUBROUTINE mp_sum_root_rv(msg,root,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m1, msglen, &
                                                taskid
    REAL(KIND=dp), ALLOCATABLE               :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_comm_rank ( gid, taskid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_rank @ mp_sum_root_rm" )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    ALLOCATE (res(m1),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_root_rv" )
    CALL mpi_reduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,&
         root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_reduce @ mp_sum_root_rv" )
    IF ( taskid == root ) THEN
       msg = res
    END IF
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_root_rv
! *****************************************************************************
  SUBROUTINE mp_sum_rm(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m1, m2, msglen
    REAL(KIND=dp), ALLOCATABLE               :: res( :, : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    ALLOCATE (res(m1,m2),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_rm" )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_rm" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_rm
! *****************************************************************************
  SUBROUTINE mp_sum_root_rm(msg,root,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m1, m2, msglen, &
                                                taskid
    REAL(KIND=dp), ALLOCATABLE               :: res( :, : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_comm_rank ( gid, taskid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_rank @ mp_sum_root_rm" )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    ALLOCATE (res(m1,m2),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_root_rm" )
    CALL mpi_reduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_reduce @ mp_sum_root_rm" )
    IF ( taskid == root ) THEN
       msg = res
    END IF
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_root_rm
! *****************************************************************************
  SUBROUTINE mp_sum_rm3(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m1, m2, m3, &
                                                msglen
    REAL(KIND=dp), ALLOCATABLE               :: res( :, :, : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    m3 = SIZE(msg,3)
    ALLOCATE (res(m1,m2,m3),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort("allocate @ mp_sum_rm3" )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_rm3" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_rm3
! *****************************************************************************
  SUBROUTINE mp_sum_root_rm3(msg,root,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_root_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, m1, m2, m3, &
                                                msglen, taskid
    REAL(KIND=dp), ALLOCATABLE               :: res( :, :, : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_comm_rank ( gid, taskid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_comm_rank @ mp_sum_root_rm3" )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    m3 = SIZE(msg,3)
    ALLOCATE (res(m1,m2,m3),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_root_rm3" )
    CALL mpi_reduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_reduce @ mp_sum_root_rm3" )
    IF ( taskid == root ) THEN
       msg = res
    END IF
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_root_rm3
! *****************************************************************************
  SUBROUTINE mp_sum_c1(msg,gid)
    COMPLEX(KIND=dp), INTENT(INOUT)          :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_c1', &
      routineP = moduleN//':'//routineN

    COMPLEX(KIND=dp)                         :: res
    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    msglen = 2
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_c1" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_c1
! *****************************************************************************
  SUBROUTINE mp_sum_cv(msg,gid)
    COMPLEX(KIND=dp), INTENT(INOUT)          :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_cv', &
      routineP = moduleN//':'//routineN

    COMPLEX(KIND=dp), ALLOCATABLE            :: res( : )
    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    IF (msglen>0) THEN
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_cv" )
    CALL mpi_allreduce(msg,res,2*msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_cv" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_cv
! *****************************************************************************
  SUBROUTINE mp_sum_cm(msg,gid)
    COMPLEX(KIND=dp), INTENT(INOUT)          :: msg( :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_cm', &
      routineP = moduleN//':'//routineN

    COMPLEX(KIND=dp), ALLOCATABLE            :: res( :, : )
    INTEGER                                  :: handle, ierr, m1, m2, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 2*SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    ALLOCATE (res(m1,m2),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_cm" )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_cm" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_cm
! *****************************************************************************
  SUBROUTINE mp_sum_cm3(msg,gid)
    COMPLEX(KIND=dp), INTENT(INOUT)          :: msg( :, :, : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_cm3', &
      routineP = moduleN//':'//routineN

    COMPLEX(KIND=dp), ALLOCATABLE            :: res( :, :, : )
    INTEGER                                  :: handle, ierr, m1, m2, m3, &
                                                msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 2*SIZE(msg)
    IF (msglen>0) THEN
    m1 = SIZE(msg,1)
    m2 = SIZE(msg,2)
    m3 = SIZE(msg,3)
    ALLOCATE (res(m1,m2,m3),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_sum_cm3" )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_SUM,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_sum_cm3" )
    msg = res
    DEALLOCATE (res)
    END IF
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*2*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_cm3

! *****************************************************************************
  SUBROUTINE mp_max_i(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 1
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_MAX,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_max_i" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_max_i
! *****************************************************************************
  SUBROUTINE mp_max_iv(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE                     :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_max_iv" )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_MAX,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_max_iv" )
    msg = res
    DEALLOCATE (res)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_max_iv
! *****************************************************************************
  SUBROUTINE mp_max_8iv(msg,gid)
    INTEGER(KIND=int_8), INTENT(INOUT)       :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_8iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE                     :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_max_8iv" )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER8,MPI_MAX,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_max_8iv" )
    msg = res
    DEALLOCATE (res)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_max_8iv

! *****************************************************************************
  SUBROUTINE mp_max_r(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    REAL(KIND=dp)                            :: res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 1
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_MAX,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_max_r" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_max_r

! *****************************************************************************
  SUBROUTINE mp_maxloc_i(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_maxloc_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE                     :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_maxloc_i" )
    CALL mpi_allreduce(msg,res,1,MPI_2INTEGER,MPI_MAXLOC,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_maxloc_i" )
    msg = res
    DEALLOCATE (res,STAT=ierr)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_maxloc_i

! *****************************************************************************
  SUBROUTINE mp_minloc_rv(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_minloc_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    REAL(KIND=dp), ALLOCATABLE               :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_min_rv" )
    CALL mpi_allreduce(msg,res,1,MPI_2DOUBLE_PRECISION,MPI_MINLOC,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_min_rv" )
    msg = res
    DEALLOCATE (res)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_minloc_rv

! *****************************************************************************
  SUBROUTINE mp_max_rv(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_max_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    REAL(KIND=dp), ALLOCATABLE               :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_max_rv" )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_MAX,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_max_rv" )
    msg = res
    DEALLOCATE (res)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_max_rv

! *****************************************************************************
  SUBROUTINE mp_min_i(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 1
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_MIN,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_min_i" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_min_i
! *****************************************************************************
  SUBROUTINE mp_min_iv(msg,gid)
    INTEGER, INTENT(INOUT)                   :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    INTEGER, ALLOCATABLE                     :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_min_iv" )
    CALL mpi_allreduce(msg,res,msglen,MPI_INTEGER,MPI_MIN,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_min_iv" )
    msg = res
    DEALLOCATE (res)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_min_iv
! *****************************************************************************
  SUBROUTINE mp_min_r(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    REAL(KIND=dp)                            :: res

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 1
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_MIN,gid, &
         ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_min_r" )
    msg = res
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_min_r
! *****************************************************************************
  SUBROUTINE mp_min_rv(msg,gid)
    REAL(KIND=dp), INTENT(INOUT)             :: msg( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_min_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen
    REAL(KIND=dp), ALLOCATABLE               :: res( : )

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    ALLOCATE (res(1:msglen),STAT=ierr)
    IF ( ierr /= 0 ) CALL mp_abort( "allocate @ mp_min_rv" )
    CALL mpi_allreduce(msg,res,msglen,MP_STD_REAL,MPI_MIN,gid, ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allreduce @ mp_min_rv" )
    msg = res
    DEALLOCATE (res)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_min_rv

  SUBROUTINE mp_scatter_iv(msg_scatter,msg,root,gid)
    INTEGER, INTENT(IN)                      :: msg_scatter(:)
    INTEGER, INTENT(OUT)                     :: msg( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_scatter_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    CALL mpi_scatter(msg_scatter,msglen,MPI_INTEGER,msg,&
         msglen,MPI_INTEGER,root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_gather @ mp_scatter_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=4,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    msg = msg_scatter
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_scatter_iv

! *****************************************************************************
  SUBROUTINE mp_gather_i(msg,msg_gather,root,gid)
    INTEGER, INTENT(IN)                      :: msg
    INTEGER, INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 1
    CALL mpi_gather(msg,msglen,MPI_INTEGER,msg_gather,&
         msglen,MPI_INTEGER,root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_gather @ mp_gather_i" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=4,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    msg_gather = msg
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_gather_i

! *****************************************************************************
  SUBROUTINE mp_gather_iv(msg,msg_gather,root,gid)
    INTEGER, INTENT(IN)                      :: msg( : )
    INTEGER, INTENT(OUT)                     :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    CALL mpi_gather(msg,msglen,MPI_INTEGER,msg_gather,&
         msglen,MPI_INTEGER,root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_gather @ mp_gather_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=4,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    msg_gather = msg
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_gather_iv

! *****************************************************************************
  SUBROUTINE mp_gather_r(msg,msg_gather,root,gid)
    REAL(KIND=dp), INTENT(IN)                :: msg
    REAL(KIND=dp), INTENT(OUT)               :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_r', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = 1
    CALL mpi_gather(msg,msglen,MP_STD_REAL,msg_gather,&
         msglen,MP_STD_REAL,root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_gather @ mp_gather_r" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=4,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    msg_gather = msg
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_gather_r

! *****************************************************************************
  SUBROUTINE mp_gather_rv(msg,msg_gather,root,gid)
    REAL(KIND=dp), INTENT(IN)                :: msg( : )
    REAL(KIND=dp), INTENT(OUT)               :: msg_gather( : )
    INTEGER, INTENT(IN)                      :: root, gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gather_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    msglen = SIZE(msg)
    CALL mpi_gather(msg,msglen,MP_STD_REAL,msg_gather,&
         msglen,MP_STD_REAL,root,gid,ierr)
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_gather @ mp_gather_rv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=4,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    msg_gather = msg
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_gather_rv

! *****************************************************************************
  SUBROUTINE mp_gatherv_iv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    INTEGER, DIMENSION(:), INTENT(IN)        :: sendbuf
    INTEGER, DIMENSION(:), INTENT(OUT)       :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, sendcount

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
    sendcount = SIZE(sendbuf)
#if defined(__parallel)
    t_start = m_walltime()
    CALL mpi_gatherv(sendbuf,sendcount,MPI_INTEGER,&
         recvbuf,recvcounts,displs,MPI_INTEGER,&
         root,comm,ierr)
    IF (ierr /= 0) CALL mp_stop(ierr,"mpi_gatherv@mp_gatherv_iv")
    t_end = m_walltime()
    CALL add_perf(perf_id=4,&
         count=1,&
         time=t_end-t_start,&
         msg_size=sendcount*int_size)
#else
    recvbuf(1+displs(1):) = sendbuf
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_gatherv_iv

! *****************************************************************************
  SUBROUTINE mp_gatherv_rv(sendbuf,recvbuf,recvcounts,displs,root,comm)

    REAL(KIND=dp), DIMENSION(:), INTENT(IN)  :: sendbuf
    REAL(KIND=dp), DIMENSION(:), INTENT(OUT) :: recvbuf
    INTEGER, DIMENSION(:), INTENT(IN)        :: recvcounts, displs
    INTEGER, INTENT(IN)                      :: root, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_gatherv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, sendcount

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    sendcount = SIZE(sendbuf)
#if defined(__parallel)
    t_start = m_walltime()
    CALL mpi_gatherv(sendbuf,sendcount,MP_STD_REAL,&
                     recvbuf,recvcounts,displs,MP_STD_REAL,&
                     root,comm,ierr)
    IF (ierr /= 0) CALL mp_stop(ierr,"mpi_gatherv@mp_gatherv_rv")
    t_end = m_walltime()
    CALL add_perf(perf_id=4,&
                  count=1,&
                  time=t_end-t_start,&
                  msg_size=sendcount*dp_size)
#else
    recvbuf(1+displs(1):) = sendbuf
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_gatherv_rv

! wrapper for MPI_ALLGATHERV subroutine
! one dimensional real arrays

! *****************************************************************************
  SUBROUTINE mp_allgather_i1(msgout,msgin,gid)
    INTEGER, INTENT(IN)                      :: msgout
    INTEGER, INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i1', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, rcount, scount

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    scount = 1
    rcount = 1
    CALL MPI_ALLGATHER(msgout, scount, MPI_INTEGER, &
                       msgin , rcount, MPI_INTEGER, &
                       gid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allgather @ mp_allgather_i1" )
#else
    msgin = msgout
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_allgather_i1

! *****************************************************************************
  SUBROUTINE mp_allgather_i12(msgout, msgin,gid)
    INTEGER, INTENT(IN)                      :: msgout(:)
    INTEGER, INTENT(OUT)                     :: msgin(:, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i12', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, rcount, scount

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    scount = SIZE (msgout(:))
    rcount = scount
    CALL MPI_ALLGATHER(msgout, scount, MPI_INTEGER, &
                       msgin , rcount, MPI_INTEGER, &
                       gid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allgather @ mp_allgather_i12" )
#else
    msgin(:,1) = msgout(:)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_allgather_i12

! *****************************************************************************
  SUBROUTINE mp_allgather_i23(msgout, msgin,gid)
    INTEGER, INTENT(IN)                      :: msgout(:,:)
    INTEGER, INTENT(OUT)                     :: msgin(:, :, :)
    INTEGER, INTENT(IN)                      :: gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgather_i23', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, rcount, scount

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    scount = SIZE (msgout(:,:))
    rcount = scount
    CALL MPI_ALLGATHER(msgout, scount, MPI_INTEGER, &
                       msgin , rcount, MPI_INTEGER, &
                       gid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allgather @ mp_allgather_i23" )
#else
    msgin(:,:,1) = msgout(:,:)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_allgather_i23

! *****************************************************************************
  SUBROUTINE mp_allgatherv_rv(msgout,msgin,rcount,rdispl,gid)
    REAL(KIND=dp), INTENT(IN)                :: msgout( : )
    REAL(KIND=dp), INTENT(OUT)               :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, scount

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    scount = SIZE ( msgout )
    CALL MPI_ALLGATHERV(msgout, scount, MP_STD_REAL, msgin, rcount, &
                        rdispl, MP_STD_REAL, gid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allgatherv @ mp_allgatherv_rv" )
#else
    msgin = msgout
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_allgatherv_rv

! *****************************************************************************
  SUBROUTINE mp_allgatherv_iv(msgout,msgin,rcount,rdispl,gid)
    INTEGER, INTENT(IN)                      :: msgout( : )
    INTEGER, INTENT(OUT)                     :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), rdispl( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allgatherv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, scount

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    scount = SIZE ( msgout )
    CALL MPI_ALLGATHERV(msgout, scount, MPI_INTEGER, msgin, rcount, &
                        rdispl, MPI_INTEGER, gid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_allgatherv @ mp_allgatherv_iv" )
#else
    msgin = msgout
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_allgatherv_iv


! wrapper for MPI_REDUCE_SCATTER subroutine
! one dimensional real arrays
! *****************************************************************************
  SUBROUTINE mp_sum_scatter_rv(msgout,msgin,rcount,gid)
    REAL(KIND=dp), INTENT(IN)                :: msgout( : )
    REAL(KIND=dp), INTENT(OUT)               :: msgin( : )
    INTEGER, INTENT(IN)                      :: rcount( : ), gid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sum_scatter_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    CALL MPI_REDUCE_SCATTER(msgout, msgin, rcount, MP_STD_REAL, MPI_SUM, &
         gid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_reduce_scatter @ mp_sum_scatter_rv" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=3,count=1,time=t_end-t_start,&
         msg_size=rcount(1)*2*reallen)
#else
    msgin = msgout
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sum_scatter_rv

! *****************************************************************************
  SUBROUTINE mp_sendrecv_rm3(msgin,dest,msgout,source,comm)
    REAL(KIND=dp), INTENT(IN)                :: msgin( :, :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(KIND=dp), INTENT(OUT)               :: msgout( :, :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen_in, &
                                                msglen_out, recv_tag, send_tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )
    msglen_in = SIZE(msgin)
    msglen_out = SIZE(msgout)
    send_tag = 0 ! cannot think of something better here, this might be dangerous
    recv_tag = 0 ! cannot think of something better here, this might be dangerous
    CALL mpi_sendrecv(msgin,msglen_in,MP_STD_REAL,dest,send_tag,msgout,&
         msglen_out,MP_STD_REAL,source,recv_tag,comm,status(1),ierr)
    ! we do not check the status
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_sendrecv @ mp_send_recv_rm3" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,msg_size=msglen_in*reallen)
    DEALLOCATE(status)
#else
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sendrecv_rm3

! *****************************************************************************
  SUBROUTINE mp_sendrecv_rm2(msgin,dest,msgout,source,comm)
    REAL(KIND=dp), INTENT(IN)                :: msgin( :, : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(KIND=dp), INTENT(OUT)               :: msgout( :, : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_rm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen, &
                                                recv_tag, send_tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )
    msglen = SIZE(msgin,1)*SIZE(msgin,2)
    send_tag = 0 ! cannot think of something better here, this might be dangerous
    recv_tag = 0 ! cannot think of something better here, this might be dangerous
    CALL mpi_sendrecv(msgin,msglen,MP_STD_REAL,dest,send_tag,msgout,&
         msglen,MP_STD_REAL,source,recv_tag,comm,status(1),ierr)
    ! we do not check the status
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_sendrecv @ mp_send_recv_rm2" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,msg_size=msglen*reallen)
    DEALLOCATE(status)
#else
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sendrecv_rm2

! *****************************************************************************
  SUBROUTINE mp_sendrecv_rv(msgin,dest,msgout,source,comm)
    REAL(KIND=dp), INTENT(IN)                :: msgin( : )
    INTEGER, INTENT(IN)                      :: dest
    REAL(KIND=dp), INTENT(OUT)               :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen_in, &
                                                msglen_out, recv_tag, send_tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )
    msglen_in = SIZE(msgin)
    msglen_out = SIZE(msgout)
    send_tag = 0 ! cannot think of something better here, this might be dangerous
    recv_tag = 0 ! cannot think of something better here, this might be dangerous
    CALL mpi_sendrecv(msgin,msglen_in,MP_STD_REAL,dest,send_tag,msgout,&
         msglen_out,MP_STD_REAL,source,recv_tag,comm,status(1),ierr)
    ! we do not check the status
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_sendrecv @ mp_send_recv_rv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,&
         msg_size=(msglen_in+msglen_out)*reallen/2)
    DEALLOCATE(status)
#else
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sendrecv_rv

! *****************************************************************************
  SUBROUTINE mp_sendrecv_iv(msgin,dest,msgout,source,comm)
    INTEGER, INTENT(IN)                      :: msgin( : ), dest
    INTEGER, INTENT(OUT)                     :: msgout( : )
    INTEGER, INTENT(IN)                      :: source, comm

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_sendrecv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, msglen_in, &
                                                msglen_out, recv_tag, send_tag
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )
    msglen_in = SIZE(msgin)
    msglen_out = SIZE(msgout)
    send_tag = 0 ! cannot think of something better here, this might be dangerous
    recv_tag = 0 ! cannot think of something better here, this might be dangerous
    CALL mpi_sendrecv(msgin,msglen_in,MPI_INTEGER,dest,send_tag,msgout,&
         msglen_out,MPI_INTEGER,source,recv_tag,comm,status(1),ierr)
    ! we do not check the status
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_sendrecv @ mp_send_recv_iv" )
    t_end = m_walltime ( )
    CALL add_perf(perf_id=7,count=1,time=t_end-t_start,&
         msg_size=(msglen_in+msglen_out)*reallen/2)
    DEALLOCATE(status)
#else
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_sendrecv_iv



! *****************************************************************************
!> \brief non blocking send of real arrays
!> \note
!>      the arguments must be pointers to be sure that we do not get a temporaries.
!>      they must point to contiguous memory
!> \par History
!>      11.2004 created [Joost VandeVondele]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_rv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    REAL(KIND=dp), DIMENSION(:), POINTER     :: msgin
    INTEGER, INTENT(IN)                      :: dest
    REAL(KIND=dp), DIMENSION(:), POINTER     :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_rv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, msglen, &
                                                my_tag
    REAL(KIND=dp)                            :: foo

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout,1)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       CALL mpi_irecv(msgout(lower1),msglen,MP_STD_REAL,source, my_tag,&
            comm,recv_request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MP_STD_REAL,source, my_tag,&
            comm,recv_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_isendrecv_rm2" )

    msglen = SIZE(msgin,1)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       CALL mpi_isend(msgin(lower1),msglen,MP_STD_REAL,dest,my_tag,&
            comm,send_request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MP_STD_REAL,dest,my_tag,&
            comm,send_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isendrecv_rm2" )

    msglen = (msglen+SIZE(msgout,1))/2.0_dp
    t_end = m_walltime ( )
    CALL add_perf(perf_id=8,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    send_request=0
    recv_request=0
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isendrecv_rv


! *****************************************************************************
!> \brief non blocking send of real arrays
!> \note
!>      the arguments must be pointers to be sure that we do not get a temporaries.
!>      they must point to contiguous memory
!> \par History
!>      12.2008 adapted from mp_isendrecv_rv by UB
! *****************************************************************************
  SUBROUTINE mp_isendrecv_iv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    INTEGER, DIMENSION(:), POINTER           :: msgin
    INTEGER, INTENT(IN)                      :: dest
    INTEGER, DIMENSION(:), POINTER           :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, msglen, &
                                                my_tag
    REAL(KIND=dp)                            :: foo

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout,1)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       CALL mpi_irecv(msgout(lower1),msglen,MPI_INTEGER,source, my_tag,&
            comm,recv_request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MPI_INTEGER,source, my_tag,&
            comm,recv_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_isendrecv_iv" )

    msglen = SIZE(msgin,1)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       CALL mpi_isend(msgin(lower1),msglen,MPI_INTEGER,dest,my_tag,&
            comm,send_request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MPI_INTEGER,dest,my_tag,&
            comm,send_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isendrecv_iv" )

    msglen = (msglen+SIZE(msgout,1))/2.0_dp
    t_end = m_walltime ( )
    CALL add_perf(perf_id=8,count=1,time=t_end-t_start,msg_size=msglen*intlen)
#else
    send_request=0
    recv_request=0
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isendrecv_iv


! *****************************************************************************
!> \brief non blocking send of int_8 arrays
!> \note
!>      the arguments must be pointers to be sure that we do not get a temporaries.
!>      they must point to contiguous memory
!> \par History
!>      11.2007 created [Manuel Guidon]
! *****************************************************************************
  SUBROUTINE mp_isendrecv_8iv(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    INTEGER(int_8), DIMENSION(:), POINTER    :: msgin
    INTEGER, INTENT(IN)                      :: dest
    INTEGER(int_8), DIMENSION(:), POINTER    :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_8iv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, msglen, &
                                                my_tag
    INTEGER(int_8)                           :: foo

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif
#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout,1)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       CALL mpi_irecv(msgout(lower1),msglen,MPI_INTEGER8,source, my_tag,&
            comm,recv_request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MPI_INTEGER8,source, my_tag,&
            comm,recv_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_isendrecv_8iv" )

    msglen = SIZE(msgin,1)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       CALL mpi_isend(msgin(lower1),msglen,MPI_INTEGER8,dest,my_tag,&
            comm,send_request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MPI_INTEGER8,dest,my_tag,&
            comm,send_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isendrecv_8iv" )

    msglen = (msglen+SIZE(msgout,1))/2.0_dp
    t_end = m_walltime ( )
    CALL add_perf(perf_id=8,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    send_request=0
    recv_request=0
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isendrecv_8iv

! *****************************************************************************
!> \brief non blocking send of real arrays 
!> \note
!>      the arguments must be pointers to be sure that we do not get a temporaries.
!>      they must point to contiguous memory
!> \par History
!>      08.2003 created [f&j]
!> \author joost & fawzi
! *****************************************************************************
  SUBROUTINE mp_isendrecv_rm2(msgin,dest,msgout,source,comm,send_request,&
       recv_request,tag)
    REAL(KIND=dp), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest
    REAL(KIND=dp), DIMENSION(:, :), POINTER  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: send_request, recv_request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isendrecv_rm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, lower2, &
                                                msglen, my_tag
    REAL(KIND=dp)                            :: foo

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout,1)*SIZE(msgout,2)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       lower2=LBOUND(msgout,2)
       CALL mpi_irecv(msgout(lower1,lower2),msglen,MP_STD_REAL,source, my_tag,&
            comm,recv_request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MP_STD_REAL,source, my_tag,&
            comm,recv_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_isendrecv_rm2" )

    msglen = SIZE(msgin,1)*SIZE(msgin,2)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       lower2=LBOUND(msgin,2)
       CALL mpi_isend(msgin(lower1,lower2),msglen,MP_STD_REAL,dest,my_tag,&
            comm,send_request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MP_STD_REAL,dest,my_tag,&
            comm,send_request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isendrecv_rm2" )

    msglen = (msglen+SIZE(msgout,1)*SIZE(msgout,2))/2.0_dp
    t_end = m_walltime ( )
    CALL add_perf(perf_id=8,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    send_request=0
    recv_request=0 
    msgout = msgin
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isendrecv_rm2

! *****************************************************************************
!> \brief non blocking send of real arrays 
!> \note
!>      the arguments must be pointers to be sure that we do not get a temporaries.
!> \author fawzi
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
! *****************************************************************************
  SUBROUTINE mp_isend_rm3(msgin,dest,comm,request,tag)
    REAL(KIND=dp), DIMENSION(:, :, :), &
      POINTER                                :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, lower2, &
                                                lower3, msglen, my_tag
    REAL(KIND=dp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgin,1)*SIZE(msgin,2)*SIZE(msgin,3)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       lower2=LBOUND(msgin,2)
       lower3=LBOUND(msgin,3)
       CALL mpi_isend(msgin(lower1,lower2,lower3),msglen,MP_STD_REAL,dest,my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MP_STD_REAL,dest,my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isend_rm2" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=11,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    ierr=1
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isend_rm3

  SUBROUTINE mp_isend_rm2(msgin,dest,comm,request,tag)
    REAL(KIND=dp), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_rm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, lower2, &
                                                msglen, my_tag
    REAL(KIND=dp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgin,1)*SIZE(msgin,2)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       lower2=LBOUND(msgin,2)
       CALL mpi_isend(msgin(lower1,lower2),msglen,MP_STD_REAL,dest,my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MP_STD_REAL,dest,my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isend_rm2" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=11,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    ierr=1
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isend_rm2

  SUBROUTINE mp_isend_cv(msgin,dest,comm,request,tag)
    COMPLEX(KIND=dp), DIMENSION(:), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, msglen, &
                                                my_tag
    REAL(KIND=dp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgin)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       CALL mpi_isend(msgin(lower1),msglen,MP_STD_COMPLEX,dest,my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MP_STD_COMPLEX,dest,my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isend_cv" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=11,count=1,time=t_end-t_start,msg_size=2*msglen*reallen)
#else
    ierr=1
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isend_cv

#if defined(__FFTSGL)
  SUBROUTINE mp_isend_rm2_lp(msgin,dest,comm,request,tag)
    REAL(KIND=sp), DIMENSION(:, :), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_rm2_lp', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, lower2, &
                                                msglen, my_tag
    REAL(KIND=sp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgin,1)*SIZE(msgin,2)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       lower2=LBOUND(msgin,2)
       CALL mpi_isend(msgin(lower1,lower2),msglen,MP_STD_REAL,dest,my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MP_STD_REAL,dest,my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isend_rm2" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=11,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    ierr=1
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isend_rm2_lp

  SUBROUTINE mp_isend_cv_lp(msgin,dest,comm,request,tag)
    COMPLEX(KIND=sp), DIMENSION(:), POINTER  :: msgin
    INTEGER, INTENT(IN)                      :: dest, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_isend_cv_lp', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, msglen, &
                                                my_tag
    REAL(KIND=sp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgin)
    IF (msglen>0) THEN
       lower1=LBOUND(msgin,1)
       CALL mpi_isend(msgin(lower1),msglen,MP_STD_COMPLEX,dest,my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_isend(foo,msglen,MP_STD_COMPLEX,dest,my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_isend @ mp_isend_cv" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=11,count=1,time=t_end-t_start,msg_size=2*msglen*reallen)
#else
    ierr=1
    CALL mp_stop( ierr, "mp_isend called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_isend_cv_lp
#endif

! *****************************************************************************
!> \brief non blocking recieve of real arrays 
!> \note
!>      the arguments must be pointers to be sure that we do not get a temporaries.
!> \author fawzi
!> \par History
!>      9.2008 added _rm3 subroutine [Iain Bethune] (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
! *****************************************************************************
  SUBROUTINE mp_irecv_rm3(msgout,source,comm,request,tag)
    REAL(KIND=dp), DIMENSION(:, :, :), &
      POINTER                                :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_rm3', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, lower2, &
                                                lower3, msglen, my_tag
    REAL(KIND=dp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout,1)*SIZE(msgout,2)*SIZE(msgout,3)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       lower2=LBOUND(msgout,2)
       lower3=LBOUND(msgout,3)
       CALL mpi_irecv(msgout(lower1,lower2,lower3),msglen,MP_STD_REAL,source, my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MP_STD_REAL,source, my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_irecv_rm3" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=12,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    CALL mp_abort( "mp_irecv called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_irecv_rm3

  SUBROUTINE mp_irecv_rm2(msgout,source,comm,request,tag)
    REAL(KIND=dp), DIMENSION(:, :), POINTER  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_rm2', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, lower2, &
                                                msglen, my_tag
    REAL(KIND=dp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout,1)*SIZE(msgout,2)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       lower2=LBOUND(msgout,2)
       CALL mpi_irecv(msgout(lower1,lower2),msglen,MP_STD_REAL,source, my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MP_STD_REAL,source, my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_irecv_rm2" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=12,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    CALL mp_abort( "mp_irecv called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_irecv_rm2

  SUBROUTINE mp_irecv_cv(msgout,source,comm,request,tag)
    COMPLEX(KIND=dp), DIMENSION(:), POINTER  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_cv', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, msglen, &
                                                my_tag
    REAL(KIND=dp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       CALL mpi_irecv(msgout(lower1),msglen,MP_STD_COMPLEX,source, my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MP_STD_COMPLEX,source, my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_irecv_cv" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=12,count=1,time=t_end-t_start,msg_size=2*msglen*reallen)
#else
    CALL mp_abort( "mp_irecv called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_irecv_cv

#if defined(__FFTSGL)
  SUBROUTINE mp_irecv_rm2_lp(msgout,source,comm,request,tag)
    REAL(KIND=sp), DIMENSION(:, :), POINTER  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_rm2_lp', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, lower2, &
                                                msglen, my_tag
    REAL(KIND=sp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout,1)*SIZE(msgout,2)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       lower2=LBOUND(msgout,2)
       CALL mpi_irecv(msgout(lower1,lower2),msglen,MP_STD_REAL,source, my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MP_STD_REAL,source, my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_irecv_rm2" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=12,count=1,time=t_end-t_start,msg_size=msglen*reallen)
#else
    CALL mp_abort( "mp_irecv called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_irecv_rm2_lp

  SUBROUTINE mp_irecv_cv_lp(msgout,source,comm,request,tag)
    COMPLEX(KIND=sp), DIMENSION(:), POINTER  :: msgout
    INTEGER, INTENT(IN)                      :: source, comm
    INTEGER, INTENT(out)                     :: request
    INTEGER, INTENT(in), OPTIONAL            :: tag

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_irecv_cv_lp', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr, lower1, msglen, &
                                                my_tag
    REAL(KIND=sp)                            :: foo(1)

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    t_start = m_walltime ( )
    my_tag = 0
    IF (PRESENT(tag)) my_tag=tag

    msglen = SIZE(msgout)
    IF (msglen>0) THEN
       lower1=LBOUND(msgout,1)
       CALL mpi_irecv(msgout(lower1),msglen,MP_STD_COMPLEX,source, my_tag,&
            comm,request,ierr)
    ELSE
       CALL mpi_irecv(foo,msglen,MP_STD_COMPLEX,source, my_tag,&
            comm,request,ierr)
    END IF
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_ircv @ mp_irecv_cv" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=12,count=1,time=t_end-t_start,msg_size=2*msglen*reallen)
#else
    CALL mp_abort( "mp_irecv called in non parallel case" )
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_irecv_cv_lp
#endif

! *****************************************************************************
!> \brief waits for completion of the given request
!> \note
!>      see isendrecv
!> \par History
!>      08.2003 created [f&j]
!> \author joost & fawzi
! *****************************************************************************
  SUBROUTINE mp_wait(request)
    INTEGER, INTENT(inout)                   :: request

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_wait', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, ierr
    INTEGER, ALLOCATABLE, DIMENSION(:)       :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    ALLOCATE(status(MPI_STATUS_SIZE))
    t_start = m_walltime ( )

    CALL mpi_wait(request,status,ierr)
    ! we do not check the status
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_wait @ mp_wait" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=9,count=1,time=t_end-t_start)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_wait

! *****************************************************************************
!> \brief waits for completion of the given requests
!> \note
!>      see isendrecv
!> \par History
!>      08.2003 created [f&j]
!> \author joost & fawzi
! *****************************************************************************
  SUBROUTINE mp_waitall(requests)
    INTEGER, DIMENSION(:), INTENT(inout)     :: requests

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_waitall', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: count, handle, ierr
    INTEGER, ALLOCATABLE, DIMENSION(:, :)    :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    count=SIZE(requests)
    ALLOCATE(status(MPI_STATUS_SIZE,count))
    t_start = m_walltime ( )

    CALL mpi_waitall(count,requests,status,ierr)
    ! we do not check the status
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_waitall @ mp_waitall" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=9,count=1,time=t_end-t_start)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_waitall

! *****************************************************************************
!> \brief waits for completion of any of the given requests
!> \par History
!>      09.2008 created 
!> \author Iain Bethune (c) The Numerical Algorithms Group (NAG) Ltd, 2008 on behalf of the HECToR project
! *****************************************************************************
  SUBROUTINE mp_waitany(requests, completed)
    INTEGER, DIMENSION(:), INTENT(inout)     :: requests
    INTEGER, INTENT(out)                     :: completed

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_waitany', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: count, handle, ierr
    INTEGER, ALLOCATABLE, DIMENSION(:, :)    :: status

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

#if defined(__parallel)
    count=SIZE(requests)
    ALLOCATE(status(MPI_STATUS_SIZE,count))
    t_start = m_walltime ( )

    CALL mpi_waitany(count,requests,completed,status,ierr)
    ! we do not check the status
    IF ( ierr /= 0 ) CALL mp_stop( ierr, "mpi_waitany @ mp_waitany" )

    t_end = m_walltime ( )
    CALL add_perf(perf_id=9,count=1,time=t_end-t_start)
    DEALLOCATE(status)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_waitany

! *****************************************************************************
!> \brief splits the given communicator in group in subgroups trying to organize
!>      them in a way that the communication within each subgroup is
!>      efficent (but not necessarily the comunication between subgroups)
!> \param comm the mpi communicator that you want to split
!> \param sub_comm the communicator for the subgroup (created, needs to be freed later)
!> \param ngroups actual number of groups
!> \param group_distribution input  : allocated with array with the nprocs entries (0 .. nprocs-1)
!> \param output the number of the group the proc belongs to (0..ngroups-1)
!> \param subgroup_min_size the minimum size of the subgroup
!> \param n_subgroups the number of subgroups wanted
!> \param group_partition n_subgroups sized array containing the number of cpus wanted per group.
!>                         should match the total number of cpus (only used if present and associated) (0..ngroups-1)
!> \note
!>      at least one of subgroup_min_size and n_subgroups is needed,
!>      the other default to the value needed to use most processors.
!>      if less cpus are present than needed for subgroup min size, n_subgroups,
!>      just one comm is created that contains all cpus
!> \par History
!>      10.2003 created [fawzi]
!>      02.2004 modified [Joost VandeVondele]
!> \author Fawzi Mohamed
! *****************************************************************************
  SUBROUTINE mp_comm_split(comm,sub_comm,ngroups, group_distribution, &
       subgroup_min_size, n_subgroups, group_partition)
    INTEGER, INTENT(in)                      :: comm
    INTEGER, INTENT(out)                     :: sub_comm, ngroups
    INTEGER, DIMENSION(:), POINTER           :: group_distribution
    INTEGER, INTENT(in), OPTIONAL            :: subgroup_min_size, n_subgroups
    INTEGER, DIMENSION(:), OPTIONAL, POINTER :: group_partition

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_comm_split', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: color, handle, i, ierr, j, k, &
                                                mepos, my_subgroup_min_size, &
                                                nnodes

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    ! actual number of groups

    IF (.NOT. PRESENT(subgroup_min_size) .AND. .NOT. PRESENT(n_subgroups)) THEN
       CALL mp_abort(routineP//" missing arguments ")
    ENDIF
    IF (PRESENT(subgroup_min_size) .AND. PRESENT(n_subgroups)) THEN
       CALL mp_abort(routineP//" too many arguments ")
    ENDIF

    CALL mp_environ(nnodes,mepos,comm)

    IF (.NOT. ASSOCIATED(group_distribution)) THEN
       CALL mp_abort(routineP//" group_distribution not associated")
    ENDIF
    IF (LBOUND(group_distribution,1) .NE. 0 .OR. &
         UBOUND(group_distribution,1).NE.nnodes-1) THEN
       CALL mp_abort(routineP//" group_distribution wrong bounds")
    ENDIF

#if defined(__parallel)
    t_start = m_walltime ( )
    IF (PRESENT(subgroup_min_size)) THEN
       IF (subgroup_min_size<0) THEN
          CALL mp_abort(routineP//" subgroup_min_size too small")
       ENDIF
       ngroups= nnodes / subgroup_min_size
       my_subgroup_min_size = subgroup_min_size
    ELSE  ! n_subgroups
       IF (n_subgroups<=0) THEN
          CALL mp_abort(routineP//" n_subgroups too small")
       ENDIF
       IF (nnodes/n_subgroups > 0) THEN ! we have a least one cpu per group
          ngroups = n_subgroups 
       ELSE ! well, only one group then
          ngroups = 1
       ENDIF
       my_subgroup_min_size = nnodes / ngroups
    ENDIF
    DO i=0,nnodes-1
       group_distribution(i)=i / my_subgroup_min_size 
       ! if part of the rest, join the last group
       IF ( group_distribution(i) >= ngroups ) group_distribution(i)=ngroups-1
    ENDDO
    ! even the user gave a partition, see if we can use it to overwrite this choice
    IF (PRESENT(group_partition)) THEN
       IF (ASSOCIATED(group_partition)) THEN
          IF (ALL(group_partition>0) .AND. SUM(group_partition).EQ.nnodes .AND. ngroups == SIZE(group_partition)) THEN
             k=0
             DO i=0,SIZE(group_partition)-1
                DO j=1,group_partition(i)
                   group_distribution(k)=i
                   k=k+1
                ENDDO
             ENDDO
          ELSE
             ! just ignore silently as we have reasonable defaults. Probably a warning would not be to bad         
          ENDIF
       ENDIF
    ENDIF
    color=group_distribution(mepos)
    CALL mpi_comm_split(comm,color,0,sub_comm,ierr)
    IF (ierr/=mpi_success) CALL mp_stop(ierr,"in "//routineP//" split")

    t_end = m_walltime ( )
    CALL add_perf(perf_id=10,count=1,time=t_end-t_start)
#else
    CALL mp_comm_dup(comm,sub_comm) 
    group_distribution(0)=0
    ngroups=1
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_comm_split

! a bunch of currently unused features
!..mp_group
! *****************************************************************************
  SUBROUTINE mp_group(group_list,group_size,base_group,groupid)

    INTEGER, INTENT(IN)                      :: group_list( : ), group_size, &
                                                base_group
    INTEGER, INTENT(OUT)                     :: groupid

    CHARACTER(len=*), PARAMETER :: routineN = 'mp_group', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: base, handle, ierr, newgroup

    ierr = 0
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN,handle)
#endif

    ierr = 0
    groupid = base_group
#if defined(__parallel)
    t_start = m_walltime ( )
    CALL mpi_comm_group ( base_group, base, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_group @ mp_group" )

    CALL mpi_group_incl ( base, group_size, group_list, newgroup, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_group_incl @ mp_group" )

    CALL mpi_comm_create ( base_group, newgroup, groupid, ierr )
    IF ( ierr /= 0 ) CALL mp_stop ( ierr, "mpi_comm_create @ mp_group" )
    CALL add_perf(perf_id=2,count=1)
    t_end = m_walltime ( )
    CALL add_perf(perf_id=1,time=t_end-t_start)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_group

! *****************************************************************************
!> \brief Creates a double precision MPI RMA window.
!> \author UB
!> \param[out] window       created window id
!> \param[in] range        window contents
!> \param[in] gid          communicator
! *****************************************************************************
  SUBROUTINE mp_win_create_d(window, range, len, gid)
    TYPE(mp_window_type), INTENT(OUT)    :: window
    REAL(KIND=dp), DIMENSION(:), POINTER :: range
    INTEGER, INTENT(IN)                  :: gid
    INTEGER, INTENT(IN), OPTIONAL        :: len
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_d', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
    INTEGER                   :: winsize, double_size
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    IF (PRESENT (len)) THEN
       winsize = len
    ELSE
       winsize = SIZE (range)
    ENDIF
    CALL MPI_TYPE_SIZE(MPI_DOUBLE_PRECISION, double_size, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_type_size @ mp_win_create_d")
    CALL MPI_WIN_CREATE(range, winsize, double_size, 0, gid, window%id, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_win_create @ mp_win_create_d")
    t_end = m_walltime()
    CALL add_perf(perf_id=20, time=t_end-t_start)
#else
    NULLIFY(window%src_i)
    window%src_d => range
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif

  END SUBROUTINE mp_win_create_d


! *****************************************************************************
!> \brief Creates an integer MPI RMA window.
!> \author UB
!> \param[out] window       created window id
!> \param[in] range        window contents
!> \param[in] gid          communicator
! *****************************************************************************
  SUBROUTINE mp_win_create_i(window, range, len, gid)
    TYPE(mp_window_type), INTENT(OUT)   :: window
    INTEGER, DIMENSION(:), POINTER      :: range
    INTEGER, INTENT(IN)       :: gid
    INTEGER, INTENT(IN), OPTIONAL       :: len
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_win_create_i', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
    INTEGER                   :: winsize, integer_size
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    IF (PRESENT (len)) THEN
       winsize = len
    ELSE
       winsize = SIZE (range)
    ENDIF
    CALL MPI_TYPE_SIZE(MPI_INTEGER, integer_size, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_type_size @ mp_win_create_i")
    CALL MPI_WIN_CREATE(range, winsize, integer_size, 0, gid, window%id, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_win_create @ mp_win_create_i")
    t_end = m_walltime()
    CALL add_perf(perf_id=20, count=1, time=t_end-t_start)
#else
    NULLIFY (window%src_d)
    window%src_i => range
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_win_create_i


! *****************************************************************************
!> \brief Destroys an MPI RMA window.
!> \author UB
!> \param[in] window       window to free and destroy
! *****************************************************************************
  SUBROUTINE mp_window_destroy(window)
    TYPE(mp_window_type), INTENT(INOUT) :: window
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_window_destroy', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    CALL MPI_WIN_FREE(window, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_win_free @ mp_window_destroy")
    t_end = m_walltime()
    CALL add_perf(perf_id=20, count=1, time=t_end-t_start)
#else
    NULLIFY(window%src_d, window%src_i)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_window_destroy


! *****************************************************************************
!> \brief Initiates an RMA access
!> \author UB
!> \param window       RMA access window
!> \param[in] remote   process to which access is initiated (target)
!> \param[in] shared   (optional) access is shared (default is exclusive)
!> \param[in] nocheck  (optional) don't check for others' locks (default is
!>                     to be careful and check)
! *****************************************************************************
  SUBROUTINE mp_window_lock(window, remote, shared, nocheck)
    TYPE(mp_window_type), INTENT(INOUT) :: window
    INTEGER, INTENT(IN)      :: remote
    LOGICAL, INTENT(IN), OPTIONAL       :: shared, nocheck
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_window_lock', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
    INTEGER                   :: exclusivity, assertion
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    exclusivity = MPI_LOCK_EXCLUSIVE
    IF (PRESENT (shared)) THEN
       IF (shared) exclusivity = MPI_LOCK_SHARED
    ENDIF
    assertion = 0
    IF (PRESENT (nocheck)) THEN
       IF (nocheck) assertion = MPI_MODE_NOCHECK
    ENDIF
    CALL MPI_WIN_LOCK (exclusivity, remote, assertion, window%id, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_win_lock @ mp_window_lock")
    t_end = m_walltime()
    CALL add_perf(perf_id=19, count=1, time=t_end-t_start)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_window_lock

! *****************************************************************************
!> \brief Terminates an RMA access
!> \author UB
!> \param[in] window   RMA access window
!> \param[in] remote   process to which access is initiated (target)
! *****************************************************************************
  SUBROUTINE mp_window_unlock(window, remote)
    TYPE(mp_window_type), INTENT(INOUT) :: window
    INTEGER, INTENT(IN)      :: remote
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_window_unlock', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    CALL MPI_WIN_UNLOCK (remote, window%id, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_win_unlock @ mp_window_unlock")
    t_end = m_walltime()
    CALL add_perf(perf_id=20, count=1, time=t_end-t_start)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_window_unlock


! *****************************************************************************
!> \brief Creates an RMA window epoch
!> \author UB
!> \param[in] window       RMA access window
!> \param[in] nothing_stored  (optional) asserts that there were no local
!>                            window updates since last sync (default is
!>                            .FALSE.)
!> \param[in] no_put_follows  (optional) asserts that nobody will put or
!>                            accumulate data in local data of window until
!>                            next fence (default is .FALSE.)
!> \param[in] no_pre_access   (optional) asserts that no RMA access calls
!>                            were made. All procesess must specify the same
!>                            flag (default is .FALSE.)
!> \param[in] no_post_access  (optional) asserts that no RMA access will be
!>                            issued. All processess must specify the same
!>                            flag (default is .FALSE.)
! *****************************************************************************
  SUBROUTINE mp_window_fence(window, nothing_stored, no_put_follows,&
       no_pre_access, no_post_access)
    TYPE(mp_window_type), INTENT(INOUT) :: window
    LOGICAL, INTENT(IN), OPTIONAL       :: nothing_stored, no_put_follows,&
                                           no_pre_access, no_post_access
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_window_unlock', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle, assertion
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    assertion = 0
    IF (PRESENT (nothing_stored)) THEN
       IF (nothing_stored) assertion = assertion + MPI_MODE_NOSTORE
    ENDIF
    IF (PRESENT (no_put_follows)) THEN
       IF (no_put_follows) assertion = assertion + MPI_MODE_NOPUT
    ENDIF
    IF (PRESENT (no_pre_access)) THEN
       IF (no_pre_access) assertion = assertion + MPI_MODE_NOPRECEDE
    ENDIF
    IF (PRESENT (no_post_access)) THEN
       IF (no_post_access) assertion = assertion + MPI_MODE_NOSUCCEED
    ENDIF
    CALL MPI_WIN_FENCE (assertion, window%id, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_win_fence @ mp_window_fence")
    t_end = m_walltime()
    CALL add_perf(perf_id=18, count=1, time=t_end-t_start)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_window_fence


! *****************************************************************************
!> \brief Fetch access using RMA, integer
!> \author UB
!> \param[in] window       RMA access window
!> \param[in] remote       remote process
!> \param[in] offset       offset in remote window (starting at 0)
!> \param[in] len          amount of data to fetch
!> \param[out] storage     integer storage of fetched data
! *****************************************************************************
  SUBROUTINE mp_rma_get_i(window, remote, offset, len, storage)
    TYPE(mp_window_type), INTENT(IN) :: window
    INTEGER, INTENT(IN)      :: remote, offset
    INTEGER, DIMENSION(:), POINTER :: storage
    INTEGER, INTENT(IN)       :: len
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rma_get_i', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    CALL MPI_GET(storage, len, MPI_INTEGER, remote, offset, len,&
         MPI_INTEGER, window%id, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_get @ mp_rma_get_i")
    t_end = m_walltime()
    CALL add_perf(perf_id=17, count=1, time=t_end-t_start,&
         msg_size=len*intlen)
#else
    storage(LBOUND(storage,1):LBOUND(storage,1)+len-1) &
         = window%src_i(LBOUND(window%src_i,1)+offset&
         : LBOUND(window%src_i,1)+offset+len-1)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_rma_get_i


! *****************************************************************************
!> \brief Fetch access using RMA, double
!> \author UB
!> \param[in] window       RMA access window
!> \param[in] remote       remote process
!> \param[in] offset       offset in remote window (starting at 0)
!> \param[in] len          amount of data to fetch
!> \param[out] storage     double precision storage of fetched data
! *****************************************************************************
  SUBROUTINE mp_rma_get_d(window, remote, offset, len, storage)
    TYPE(mp_window_type), INTENT(IN) :: window
    INTEGER, INTENT(IN)      :: remote, offset
    REAL(KIND=dp), DIMENSION(:), POINTER :: storage
    INTEGER, INTENT(IN)       :: len
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rma_get_d', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    CALL MPI_GET(storage, len, MPI_DOUBLE_PRECISION, remote, offset, len,&
         MPI_DOUBLE_PRECISION, window%id, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_get @ mp_rma_get_d")
    t_end = m_walltime()
    CALL add_perf(perf_id=17, count=1,time=t_end-t_start,&
         msg_size=len*reallen)
#else
    storage(LBOUND(storage,1):LBOUND(storage,1)+len-1) =&
         window%src_d(LBOUND(window%src_d,1)+offset&
         : LBOUND(window%src_d,1)+offset+len-1)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_rma_get_d


! *****************************************************************************
!> \brief Store access using RMA, integer
!> \author UB
!> \param[in] window       RMA access window
!> \param[in] storage      integer data to store
!> \param[in] len          amount of data to store
!> \param[in] remote       remote process
!> \param[in] offset       offset in remote window (starting at 0)
! *****************************************************************************
  SUBROUTINE mp_rma_put_i(window, storage, len, remote, offset)
    TYPE(mp_window_type), INTENT(IN)  :: window
    INTEGER, INTENT(IN)               :: remote, offset
    INTEGER, DIMENSION(:), POINTER    :: storage
    INTEGER, INTENT(IN)         :: len
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rma_put_i', &
         routineP = moduleN//':'//routineN
    INTEGER                        :: ierr, handle
    INTEGER, DIMENSION(:), POINTER :: win_data
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    CALL MPI_PUT(storage, len, MPI_INTEGER, remote, offset, len, MPI_INTEGER,&
         window, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_put @ mp_rma_put_i")
    t_end = m_walltime()
    CALL add_perf(perf_id=16, count=1, time=t_end-t_start,&
         msg_size=intlen*len)
#else
    win_data => window%src_i
    win_data(LBOUND(window%src_i,1)+offset &
         : LBOUND(window%src_i,1)+offset+len-1) =&
    storage(LBOUND(storage,1):LBOUND(storage,1)+len-1)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_rma_put_i


! *****************************************************************************
!> \brief Store access using RMA, doubles
!> \author UB
!> \param[in] window       RMA access window
!> \param[in] storage      double precision data to store
!> \param[in] len          amount of data to store
!> \param[in] remote       remote process
!> \param[in] offset       offset in remote window (starting at 0)
! *****************************************************************************
  SUBROUTINE mp_rma_put_d(window, storage, len, remote, offset)
    TYPE(mp_window_type), INTENT(IN)      :: window
    INTEGER, INTENT(IN)                   :: remote, offset
    REAL(KIND=dp), DIMENSION(:), POINTER  :: storage
    INTEGER, INTENT(IN)         :: len
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_rma_put_d', &
         routineP = moduleN//':'//routineN
    INTEGER                     :: ierr, handle
    REAL(KIND=dp), DIMENSION(:), POINTER     :: win_data
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    CALL MPI_PUT(storage, len, MPI_DOUBLE_PRECISION, remote, offset, len,&
         MPI_DOUBLE_PRECISION, window, ierr)
    IF (ierr /= 0) CALL mp_stop(ierr, "mpi_put @ mp_rma_put_d")
    t_end = m_walltime()
    CALL add_perf(perf_id=16, count=1, time=t_end-t_start,&
         msg_size=len*reallen)
#else
    win_data => window%src_d
    win_data(LBOUND(window%src_d,1)+offset&
         : LBOUND(window%src_d,1)+offset+len-1) =&
    storage(LBOUND(storage,1):LBOUND(storage,1)+len-1)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_rma_put_d


! *****************************************************************************
!> \brief Allocates special parallel memory, integer version
!> \author UB
!> \param[in]  data      pointer to integer array to allocate
!> \param[in]  len       number of integers to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_allocate_i(DATA, len, stat)
    INTEGER, DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(IN)                 :: len
    INTEGER, INTENT(OUT), OPTIONAL      :: stat
    INTEGER                   :: ierr, handle
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_i', &
         routineP = moduleN//':'//routineN
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    NULLIFY(DATA)
    !WRITE(*,*)routineN//' Allocating size len=',len
    CALL mp_alloc_mem(DATA, len, stat=ierr)
    IF (PRESENT (stat)) THEN
       stat = ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "mpi_alloc_mem @ mp_allocate_i")
    ENDIF
    !WRITE(*,*)routineN//' Allocated size ',SIZE(data)
    DO ierr = 1, len
       DATA(ierr) = -333
    ENDDO
    t_end = m_walltime()
    CALL add_perf(perf_id=15, count=1, time=t_end-t_start)
#else
    ALLOCATE(DATA(len), stat=ierr)
    IF (PRESENT (stat)) THEN
       stat = ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "ALLOCATE @ mp_allocate_i")
    ENDIF
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_allocate_i


! *****************************************************************************
!> \brief Allocates special parallel memory, double precision version
!> \author UB
!> \param[in]  data      pointer to double precision array to allocate
!> \param[in]  len       number of doubles to allocate
!> \param[out] stat      (optional) allocation status result
! *****************************************************************************
  SUBROUTINE mp_allocate_d(DATA, len, stat)
    REAL(KIND=dp), DIMENSION(:), POINTER  :: DATA
    INTEGER, INTENT(IN)                   :: len
    INTEGER, INTENT(OUT), OPTIONAL        :: stat
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_allocate_d', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    NULLIFY(DATA)
    !WRITE(*,*)routineN//' Allocating size len=',len
    CALL mp_alloc_mem(DATA, len, stat=ierr)
    IF (PRESENT (stat)) THEN
       stat = ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "mpi_alloc_mem @ mp_allocate_d")
    ENDIF
    !WRITE(*,*)routineN//' Allocated size ',SIZE(data)
    DO ierr = 1, SIZE(DATA)
       DATA(ierr) = -333
    ENDDO
    t_end = m_walltime()
    CALL add_perf(perf_id=15, count=1, time=t_end-t_start)
#else
    ALLOCATE(DATA(len), stat=ierr)
    IF (PRESENT (stat)) THEN
       stat = ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "ALLOCATE @ mp_allocate_d")
    ENDIF
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_allocate_d


! *****************************************************************************
!> \brief Deallocates special parallel memory, integer version
!> \author UB
!> \param[in] data         pointer to special memory to deallocate
! *****************************************************************************
  SUBROUTINE mp_deallocate_i(DATA, stat)
    INTEGER, DIMENSION(:), POINTER      :: DATA
    INTEGER, INTENT(OUT), OPTIONAL      :: stat
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_i', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    !WRITE(*,*)routineN//' Deallocating', SIZE(data)
    CALL mp_free_mem(DATA, ierr)
    IF (PRESENT (stat)) THEN
       stat = ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "mpi_free_mem @ mp_deallocate_i")
    ENDIF
    NULLIFY(DATA)
    t_end = m_walltime()
    CALL add_perf(perf_id=15, count=1, time=t_end-t_start)
#else
    DEALLOCATE(DATA, stat=ierr)
    IF (PRESENT (stat)) THEN
       stat=ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "DEALLOCATE @ mp_deallocate_i")
    ENDIF
    NULLIFY(DATA)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_deallocate_i


! *****************************************************************************
!> \brief Deallocates special parallel memory, double version
!> \author UB
!> \param[in] data         pointer to special memory to deallocate
! *****************************************************************************
  SUBROUTINE mp_deallocate_d(DATA, stat)
    REAL(KIND=dp), DIMENSION(:), POINTER     :: DATA
    INTEGER, INTENT(OUT), OPTIONAL           :: stat
    CHARACTER(len=*), PARAMETER :: routineN = 'mp_deallocate_d', &
         routineP = moduleN//':'//routineN
    INTEGER                   :: ierr, handle
!   ---------------------------------------------------------------------------
#if defined(__mp_timeset__)
    CALL timeset_mp(routineN, handle)
#endif
#if defined(__parallel)
    t_start = m_walltime()
    ierr = 0
    !WRITE(*,*)routineN//' Deallocating', SIZE(data)
    CALL mp_free_mem(DATA, ierr)
    IF (PRESENT (stat)) THEN
       stat = ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "mpi_free_mem @ mp_deallocate_d")
    ENDIF
    NULLIFY(DATA)
    t_end = m_walltime()
    CALL add_perf(perf_id=15, count=1, time=t_end-t_start)
#else
    DEALLOCATE(DATA, stat=ierr)
    IF (PRESENT (stat)) THEN
       stat = ierr
    ELSE
       IF (ierr /= 0) CALL mp_stop(ierr, "DEALLOCATE @ mp_deallocate_d")
    ENDIF
    NULLIFY(DATA)
#endif
#if defined(__mp_timeset__)
    CALL timestop_mp(handle)
#endif
  END SUBROUTINE mp_deallocate_d


END MODULE message_passing
